import datetime
import typing
from functools import partial
from abc import ABCMeta, abstractmethod

from torchqtm.data.data_portal import DataPortal
from torchqtm.finance.account import Account
import empyrical as ep
import numpy as np
import pandas as pd

# 这里我不是很想去写这个, 因为目前参数命名不规范, 我后边自己再改改
# 另外我感觉使用handle_event可能比他更好


class StatsPacket(object):
    def __init__(self):
        self.cumulative_risk_metrics = {}


class Metric(object, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def start_of_simulation(
            cls,
            account: Account,
            emission_rate: typing.Literal["daily", "minute"],
            trading_calendar,
            sessions: pd.DatetimeIndex,
            benchmark_source,
    ):
        pass

    @classmethod
    def end_of_simulation(
            cls,
            packet: typing.Dict,
            account: Account,
            trading_calendar,
            sessions: pd.DatetimeIndex,
            data_portal: DataPortal,
            benchmark_source,
    ):
        pass

    @classmethod
    def start_of_session(
            cls,
            account: Account,
            session: pd.Timestamp,
            data_portal: DataPortal,
    ):
        pass

    @classmethod
    def end_of_session(
            cls,
            packet: typing.Dict,
            account: Account,
            session: pd.Timestamp,
            session_idx: int,
            data_portal: DataPortal,
    ):
        pass

    @classmethod
    def start_of_bar(
            cls,
            packet: typing.Dict,
            account: Account,
            dt: pd.Timestamp,
            session_idx: int,
            data_portal: DataPortal,
    ):
        pass

    @classmethod
    def end_of_bar(
            cls,
            packet: typing.Dict,
            account: Account,
            dt: pd.Timestamp,
            session_idx: int,
            data_portal: DataPortal,
    ):
        pass


class Returns(Metric):
    @classmethod
    def end_of_bar(cls, packet, account, dt, session_idx, data_portal):
        packet["minute_perf"]["returns"] = account.todays_returns
        packet["cumulative_perf"]["returns"] = account.ledger.returns
        packet["cumulative_risk_metrics"]["algorithm_period_return"] = account.ledger.returns

    @classmethod
    def end_of_session(cls, packet, account, session, session_idx, data_portal):
        packet["daily_perf"]["returns"] = account.todays_returns
        packet["cumulative_perf"]["returns"] = account.ledger.returns
        packet["cumulative_risk_metrics"]["algorithm_period_return"] = account.ledger.returns


class PNL(Metric):
    """Tracks daily and cumulative PNL."""
    _previous_pnl: float = 0.0

    @classmethod
    def start_of_simulation(cls, account, emission_rate, trading_calendar, sessions, benchmark_source):
        cls._previous_pnl = 0.0

    @classmethod
    def start_of_session(cls, account: Account, session, data_portal):
        cls._previous_pnl = account.ledger.pnl

    @classmethod
    def _end_of_period(cls, field, packet, account):
        pnl = account.ledger.pnl
        packet[field]["pnl"] = pnl - cls._previous_pnl
        packet["cumulative_perf"]["pnl"] = account.ledger.pnl

    @classmethod
    def end_of_bar(cls, packet, account, dt, session_idx, data_portal):
        pnl = account.ledger.pnl
        packet["minute_perf"]["pnl"] = pnl - cls._previous_pnl
        packet["cumulative_perf"]["pnl"] = account.ledger.pnl

    @classmethod
    def end_of_session(cls, packet, account, session, session_idx, data_portal):
        pnl = account.ledger.pnl
        packet["daily_perf"]["pnl"] = pnl - cls._previous_pnl
        packet["cumulative_perf"]["pnl"] = account.ledger.pnl


class Orders(Metric):

    def end_of_bar(self, packet, account, dt, session_idx, data_portal):
        packet["minute_perf"]["orders"] = account.get_orders(dt)

    def end_of_session(self, packet, account, dt, session_idx, data_portal):
        packet["daily_perf"]["orders"] = account.get_orders()


class Transactions(Metric):
    """Tracks daily transactions."""

    def end_of_bar(self, packet, account, dt, session_idx, data_portal):
        packet["minute_perf"]["transactions"] = account.get_transactions(dt)

    def end_of_session(self, packet, account, dt, session_idx, data_portal):
        packet["daily_perf"]["transactions"] = account.get_transactions()


# class Positions(Metric):
#
#     def end_of_bar(
#             self,
#             packet: typing.Dict,
#             account: Account,
#             dt: pd.Timestamp,
#             session_idx: int,
#             data_portal: DataPortal,
#     ):
#         packet["minute_perf"]["positions"] = account.get_positions(dt)
#
#     def end_of_session(
#             self,
#             packet: typing.Dict,
#             account: Account,
#             dt: pd.Timestamp,
#             session_idx: int,
#             data_portal: DataPortal,
#     ):
#         packet["daily_perf"]["positions"] = account.get_positions()


# 这个类看起来不太好, 这个操纵明显可以通过向量化完成
class ReturnsStatistics(Metric):
    """A metric that reports an end of simulation scalar or time series
    computed from the algorithm returns

    Parameters
    ----------
    function: callable
        A function that accepts a numpy array of returns and returns a scalar
        or time series
    filed_name: str, optional
        The name of the field. If not provided, it will be
        ``function.__name__``.
    """
    def __init__(self, function: typing.Callable, filed_name: str):
        super().__init__()
        if filed_name is None:
            filed_name = function.__name__

        self._function = function
        self._filed_name = filed_name

    def end_of_bar(self, packet, account, dt, session_idx, data_portal):
        res = self._function(account.daily_returns_series.values[: session_idx + 1])
        if not np.isfinite(res):
            res = None
        packet["cumulative_risk_metrics"][self._filed_name] = res

    def end_of_session(self, packet, account, dt, session_idx, data_portal):
        res = self._function(account.daily_returns_series.values[: session_idx + 1])
        if not np.isfinite(res):
            res = None
        packet["cumulative_risk_metrics"][self._filed_name] = res


class AlphaBeta(Metric):
    def start_of_simulation(
            self,
            account: Account,
            emission_rate: typing.Literal["daily", "minute"],
            trading_calendar,
            sessions: pd.DatetimeIndex,
            benchmark_source,
    ):
        pass


class NumTradingDays(Metric):

    _num_trading_days: int = 0

    def start_of_simulation(
            self,
            account: Account,
            emission_rate: typing.Literal["daily", "minute"],
            trading_calendar,
            sessions: pd.DatetimeIndex,
            benchmark_source,
    ):
        self._num_trading_days = 0

    def start_of_session(
            self,
            account: Account,
            session: pd.Timestamp,
            data_portal: DataPortal,
    ):
        self._num_trading_days += 1

    def end_of_bar(
            self,
            packet: typing.Dict,
            account: Account,
            dt: pd.Timestamp,
            session_idx: int,
            data_portal: DataPortal,
    ):
        packet["cumulative_risk_metrics"]["trading_days"] = self._num_trading_days








import pandas as pd
from typing import Optional, Union, Iterable, Dict, Hashable


class TqFrame(pd.DataFrame):
    @property
    def _constructor(self):
        return TqFrame

    def get_rows(self, row_index: Union[int, Iterable[int]]):
        return self.iloc[row_index]





from torchqtm.utils.exchange_calendar import XSHGExchangeCalendar, DateIterator


class BarData:
    """
    DataPortal的外部封装
    """
    def __init__(
            self,
            data_portal,
            simulation_dt_func,
            data_frequency,
            trading_calendar: XSHGExchangeCalendar,
            restrictions,
    ):
        self.data_portal = data_portal
        self.simulation_dt_func = simulation_dt_func
        self.data_frequency = data_frequency
        self._views = {}
        self._daily_mode = (self.data_frequency == "daily")
        self._adjust_minutes = False
        self._trading_calendar = trading_calendar
        self._is_restricted = restrictions.is_restricted
        self._date_iterator = DateIterator(start_date=trading_calendar.start_date)

    def update(self):
        self._date_iterator.update()

    @property
    def current_session(self):
        return self._date_iterator.current_session

    def _get_current_minute(self):
        pass

    def current(self, asset, filed):
        return self._

    def current_chain(self, continues_future):
        pass

    def can_trade(self, assets):
        pass

    def _can_trade_for_asset(self, asset, dt, adjusted_dt, data_portal):
        pass

    def is_stale(self, assets):
        pass

    def _is_stale_for_asset(self, asset, dt, adjusted_dt, data_portal):
        pass

    def history(self, assets, fields, bar_count, frequency):
        pass



    






# from numexpr import evaluate
# import numpy as np
# from numpy import broadcast_arrays
# from scipy.stats import (
#     linregress,
#     spearmanr,
# )
#
# from zipline.assets import Asset
# from zipline.errors import IncompatibleTerms
# from zipline.pipeline.factors import CustomFactor
# from zipline.pipeline.filters import SingleAsset
# from zipline.pipeline.mixins import StandardOutputs
# from zipline.pipeline.sentinels import NotSpecified
# from zipline.pipeline.term import AssetExists
# from zipline.utils.input_validation import (
#     expect_bounded,
#     expect_dtypes,
#     expect_types,
# )
# from zipline.utils.math_utils import nanmean
# from zipline.utils.numpy_utils import (
#     float64_dtype,
#     int64_dtype,
# )
#
#
# from .basic import Returns
#
#
# ALLOWED_DTYPES = (float64_dtype, int64_dtype)
#
#
# class _RollingCorrelation(CustomFactor):
#
#     @expect_dtypes(base_factor=ALLOWED_DTYPES, target=ALLOWED_DTYPES)
#     @expect_bounded(correlation_length=(2, None))
#     def __new__(cls,
#                 base_factor,
#                 target,
#                 correlation_length,
#                 mask=NotSpecified):
#         if target.ndim == 2 and base_factor.mask is not target.mask:
#             raise IncompatibleTerms(term_1=base_factor, term_2=target)
#
#         return super(_RollingCorrelation, cls).__new__(
#             cls,
#             inputs=[base_factor, target],
#             window_length=correlation_length,
#             mask=mask,
#         )
#
#
# class RollingPearson(_RollingCorrelation):
#     """
#     A Factor that computes pearson correlation coefficients between the columns
#     of a given Factor and either the columns of another Factor/BoundColumn or a
#     slice/single column of data.
#
#     Parameters
#     ----------
#     base_factor : zipline.pipeline.Factor
#         The factor for which to compute correlations of each of its columns
#         with `target`.
#     target : zipline.pipeline.Term with a numeric dtype
#         The term with which to compute correlations against each column of data
#         produced by `base_factor`. This term may be a Factor, a BoundColumn or
#         a Slice. If `target` is two-dimensional, correlations are computed
#         asset-wise.
#     correlation_length : int
#         Length of the lookback window over which to compute each correlation
#         coefficient.
#     mask : zipline.pipeline.Filter, optional
#         A Filter describing which assets (columns) of `base_factor` should have
#         their correlation with `target` computed each day.
#
#     See Also
#     --------
#     :func:`scipy.stats.pearsonr`
#     :meth:`Factor.pearsonr`
#     :class:`zipline.pipeline.factors.RollingPearsonOfReturns`
#
#     Notes
#     -----
#     Most users should call Factor.pearsonr rather than directly construct an
#     instance of this class.
#     """
#     window_safe = True
#
#     def compute(self, today, assets, out, base_data, target_data):
#         vectorized_pearson_r(
#             base_data,
#             target_data,
#             allowed_missing=0,
#             out=out,
#         )
#
#
# class RollingSpearman(_RollingCorrelation):
#     """
#     A Factor that computes spearman rank correlation coefficients between the
#     columns of a given Factor and either the columns of another
#     Factor/BoundColumn or a slice/single column of data.
#
#     Parameters
#     ----------
#     base_factor : zipline.pipeline.Factor
#         The factor for which to compute correlations of each of its columns
#         with `target`.
#     target : zipline.pipeline.Term with a numeric dtype
#         The term with which to compute correlations against each column of data
#         produced by `base_factor`. This term may be a Factor, a BoundColumn or
#         a Slice. If `target` is two-dimensional, correlations are computed
#         asset-wise.
#     correlation_length : int
#         Length of the lookback window over which to compute each correlation
#         coefficient.
#     mask : zipline.pipeline.Filter, optional
#         A Filter describing which assets (columns) of `base_factor` should have
#         their correlation with `target` computed each day.
#
#     See Also
#     --------
#     :func:`scipy.stats.spearmanr`
#     :meth:`Factor.spearmanr`
#     :class:`zipline.pipeline.factors.RollingSpearmanOfReturns`
#
#     Notes
#     -----
#     Most users should call Factor.spearmanr rather than directly construct an
#     instance of this class.
#     """
#     window_safe = True
#
#     def compute(self, today, assets, out, base_data, target_data):
#         # If `target_data` is a Slice or single column of data, broadcast it
#         # out to the same shape as `base_data`, then compute column-wise. This
#         # is efficient because each column of the broadcasted array only refers
#         # to a single memory location.
#         target_data = broadcast_arrays(target_data, base_data)[0]
#         for i in range(len(out)):
#             out[i] = spearmanr(base_data[:, i], target_data[:, i])[0]
#
#
# class RollingLinearRegression(CustomFactor):
#     """
#     A Factor that performs an ordinary least-squares regression predicting the
#     columns of a given Factor from either the columns of another
#     Factor/BoundColumn or a slice/single column of data.
#
#     Parameters
#     ----------
#     dependent : zipline.pipeline.Factor
#         The factor whose columns are the predicted/dependent variable of each
#         regression with `independent`.
#     independent : zipline.pipeline.slice.Slice or zipline.pipeline.Factor
#         The factor/slice whose columns are the predictor/independent variable
#         of each regression with `dependent`. If `independent` is a Factor,
#         regressions are computed asset-wise.
#     regression_length : int
#         Length of the lookback window over which to compute each regression.
#     mask : zipline.pipeline.Filter, optional
#         A Filter describing which assets (columns) of `dependent` should be
#         regressed against `independent` each day.
#
#     See Also
#     --------
#     :func:`scipy.stats.linregress`
#     :meth:`Factor.linear_regression`
#     :class:`zipline.pipeline.factors.RollingLinearRegressionOfReturns`
#
#     Notes
#     -----
#     Most users should call Factor.linear_regression rather than directly
#     construct an instance of this class.
#     """
#     outputs = ['alpha', 'beta', 'r_value', 'p_value', 'stderr']
#
#     @expect_dtypes(dependent=ALLOWED_DTYPES, independent=ALLOWED_DTYPES)
#     @expect_bounded(regression_length=(2, None))
#     def __new__(cls,
#                 dependent,
#                 independent,
#                 regression_length,
#                 mask=NotSpecified):
#         if independent.ndim == 2 and dependent.mask is not independent.mask:
#             raise IncompatibleTerms(term_1=dependent, term_2=independent)
#
#         return super(RollingLinearRegression, cls).__new__(
#             cls,
#             inputs=[dependent, independent],
#             window_length=regression_length,
#             mask=mask,
#         )
#
#     def compute(self, today, assets, out, dependent, independent):
#         alpha = out.alpha
#         beta = out.beta
#         r_value = out.r_value
#         p_value = out.p_value
#         stderr = out.stderr
#
#         def regress(y, x):
#             regr_results = linregress(y=y, x=x)
#             # `linregress` returns its results in the following order:
#             # slope, intercept, r-value, p-value, stderr
#             alpha[i] = regr_results[1]
#             beta[i] = regr_results[0]
#             r_value[i] = regr_results[2]
#             p_value[i] = regr_results[3]
#             stderr[i] = regr_results[4]
#
#         # If `independent` is a Slice or single column of data, broadcast it
#         # out to the same shape as `dependent`, then compute column-wise. This
#         # is efficient because each column of the broadcasted array only refers
#         # to a single memory location.
#         independent = broadcast_arrays(independent, dependent)[0]
#         for i in range(len(out)):
#             regress(y=dependent[:, i], x=independent[:, i])
#
#
# class RollingPearsonOfReturns(RollingPearson):
#     """
#     Calculates the Pearson product-moment correlation coefficient of the
#     returns of the given asset with the returns of all other assets.
#
#     Pearson correlation is what most people mean when they say "correlation
#     coefficient" or "R-value".
#
#     Parameters
#     ----------
#     target : zipline.assets.Asset
#         The asset to correlate with all other assets.
#     returns_length : int >= 2
#         Length of the lookback window over which to compute returns. Daily
#         returns require a window length of 2.
#     correlation_length : int >= 1
#         Length of the lookback window over which to compute each correlation
#         coefficient.
#     mask : zipline.pipeline.Filter, optional
#         A Filter describing which assets should have their correlation with the
#         target asset computed each day.
#
#     Notes
#     -----
#     Computing this factor over many assets can be time consuming. It is
#     recommended that a mask be used in order to limit the number of assets over
#     which correlations are computed.
#
#     Examples
#     --------
#     Let the following be example 10-day returns for three different assets::
#
#                        SPY    MSFT     FB
#         2017-03-13    -.03     .03    .04
#         2017-03-14    -.02    -.03    .02
#         2017-03-15    -.01     .02    .01
#         2017-03-16       0    -.02    .01
#         2017-03-17     .01     .04   -.01
#         2017-03-20     .02    -.03   -.02
#         2017-03-21     .03     .01   -.02
#         2017-03-22     .04    -.02   -.02
#
#     Suppose we are interested in SPY's rolling returns correlation with each
#     stock from 2017-03-17 to 2017-03-22, using a 5-day look back window (that
#     is, we calculate each correlation coefficient over 5 days of data). We can
#     achieve this by doing::
#
#         rolling_correlations = RollingPearsonOfReturns(
#             target=sid(8554),
#             returns_length=10,
#             correlation_length=5,
#         )
#
#     The result of computing ``rolling_correlations`` from 2017-03-17 to
#     2017-03-22 gives::
#
#                        SPY   MSFT     FB
#         2017-03-17       1    .15   -.96
#         2017-03-20       1    .10   -.96
#         2017-03-21       1   -.16   -.94
#         2017-03-22       1   -.16   -.85
#
#     Note that the column for SPY is all 1's, as the correlation of any data
#     series with itself is always 1. To understand how each of the other values
#     were calculated, take for example the .15 in MSFT's column. This is the
#     correlation coefficient between SPY's returns looking back from 2017-03-17
#     (-.03, -.02, -.01, 0, .01) and MSFT's returns (.03, -.03, .02, -.02, .04).
#
#     See Also
#     --------
#     :class:`zipline.pipeline.factors.RollingSpearmanOfReturns`
#     :class:`zipline.pipeline.factors.RollingLinearRegressionOfReturns`
#     """
#     def __new__(cls,
#                 target,
#                 returns_length,
#                 correlation_length,
#                 mask=NotSpecified):
#         # Use the `SingleAsset` filter here because it protects against
#         # inputting a non-existent target asset.
#         returns = Returns(
#             window_length=returns_length,
#             mask=(AssetExists() | SingleAsset(asset=target)),
#         )
#         return super(RollingPearsonOfReturns, cls).__new__(
#             cls,
#             base_factor=returns,
#             target=returns[target],
#             correlation_length=correlation_length,
#             mask=mask,
#         )
#
#
# class RollingSpearmanOfReturns(RollingSpearman):
#     """
#     Calculates the Spearman rank correlation coefficient of the returns of the
#     given asset with the returns of all other assets.
#
#     Parameters
#     ----------
#     target : zipline.assets.Asset
#         The asset to correlate with all other assets.
#     returns_length : int >= 2
#         Length of the lookback window over which to compute returns. Daily
#         returns require a window length of 2.
#     correlation_length : int >= 1
#         Length of the lookback window over which to compute each correlation
#         coefficient.
#     mask : zipline.pipeline.Filter, optional
#         A Filter describing which assets should have their correlation with the
#         target asset computed each day.
#
#     Notes
#     -----
#     Computing this factor over many assets can be time consuming. It is
#     recommended that a mask be used in order to limit the number of assets over
#     which correlations are computed.
#
#     See Also
#     --------
#     :class:`zipline.pipeline.factors.RollingPearsonOfReturns`
#     :class:`zipline.pipeline.factors.RollingLinearRegressionOfReturns`
#     """
#     def __new__(cls,
#                 target,
#                 returns_length,
#                 correlation_length,
#                 mask=NotSpecified):
#         # Use the `SingleAsset` filter here because it protects against
#         # inputting a non-existent target asset.
#         returns = Returns(
#             window_length=returns_length,
#             mask=(AssetExists() | SingleAsset(asset=target)),
#         )
#         return super(RollingSpearmanOfReturns, cls).__new__(
#             cls,
#             base_factor=returns,
#             target=returns[target],
#             correlation_length=correlation_length,
#             mask=mask,
#         )
#
#
# class RollingLinearRegressionOfReturns(RollingLinearRegression):
#     """
#     Perform an ordinary least-squares regression predicting the returns of all
#     other assets on the given asset.
#
#     Parameters
#     ----------
#     target : zipline.assets.Asset
#         The asset to regress against all other assets.
#     returns_length : int >= 2
#         Length of the lookback window over which to compute returns. Daily
#         returns require a window length of 2.
#     regression_length : int >= 1
#         Length of the lookback window over which to compute each regression.
#     mask : zipline.pipeline.Filter, optional
#         A Filter describing which assets should be regressed against the target
#         asset each day.
#
#     Notes
#     -----
#     Computing this factor over many assets can be time consuming. It is
#     recommended that a mask be used in order to limit the number of assets over
#     which regressions are computed.
#
#     This factor is designed to return five outputs:
#
#     - alpha, a factor that computes the intercepts of each regression.
#     - beta, a factor that computes the slopes of each regression.
#     - r_value, a factor that computes the correlation coefficient of each
#       regression.
#     - p_value, a factor that computes, for each regression, the two-sided
#       p-value for a hypothesis test whose null hypothesis is that the slope is
#       zero.
#     - stderr, a factor that computes the standard error of the estimate of each
#       regression.
#
#     For more help on factors with multiple outputs, see
#     :class:`zipline.pipeline.CustomFactor`.
#
#     Examples
#     --------
#     Let the following be example 10-day returns for three different assets::
#
#                        SPY    MSFT     FB
#         2017-03-13    -.03     .03    .04
#         2017-03-14    -.02    -.03    .02
#         2017-03-15    -.01     .02    .01
#         2017-03-16       0    -.02    .01
#         2017-03-17     .01     .04   -.01
#         2017-03-20     .02    -.03   -.02
#         2017-03-21     .03     .01   -.02
#         2017-03-22     .04    -.02   -.02
#
#     Suppose we are interested in predicting each stock's returns from SPY's
#     over rolling 5-day look back windows. We can compute rolling regression
#     coefficients (alpha and beta) from 2017-03-17 to 2017-03-22 by doing::
#
#         regression_factor = RollingRegressionOfReturns(
#             target=sid(8554),
#             returns_length=10,
#             regression_length=5,
#         )
#         alpha = regression_factor.alpha
#         beta = regression_factor.beta
#
#     The result of computing ``alpha`` from 2017-03-17 to 2017-03-22 gives::
#
#                        SPY    MSFT     FB
#         2017-03-17       0    .011   .003
#         2017-03-20       0   -.004   .004
#         2017-03-21       0    .007   .006
#         2017-03-22       0    .002   .008
#
#     And the result of computing ``beta`` from 2017-03-17 to 2017-03-22 gives::
#
#                        SPY    MSFT     FB
#         2017-03-17       1      .3   -1.1
#         2017-03-20       1      .2     -1
#         2017-03-21       1     -.3     -1
#         2017-03-22       1     -.3    -.9
#
#     Note that SPY's column for alpha is all 0's and for beta is all 1's, as the
#     regression line of SPY with itself is simply the function y = x.
#
#     To understand how each of the other values were calculated, take for
#     example MSFT's ``alpha`` and ``beta`` values on 2017-03-17 (.011 and .3,
#     respectively). These values are the result of running a linear regression
#     predicting MSFT's returns from SPY's returns, using values starting at
#     2017-03-17 and looking back 5 days. That is, the regression was run with
#     x = [-.03, -.02, -.01, 0, .01] and y = [.03, -.03, .02, -.02, .04], and it
#     produced a slope of .3 and an intercept of .011.
#
#     See Also
#     --------
#     :class:`zipline.pipeline.factors.RollingPearsonOfReturns`
#     :class:`zipline.pipeline.factors.RollingSpearmanOfReturns`
#     """
#     window_safe = True
#
#     def __new__(cls,
#                 target,
#                 returns_length,
#                 regression_length,
#                 mask=NotSpecified):
#         # Use the `SingleAsset` filter here because it protects against
#         # inputting a non-existent target asset.
#         returns = Returns(
#             window_length=returns_length,
#             mask=(AssetExists() | SingleAsset(asset=target)),
#         )
#         return super(RollingLinearRegressionOfReturns, cls).__new__(
#             cls,
#             dependent=returns,
#             independent=returns[target],
#             regression_length=regression_length,
#             mask=mask,
#         )
#
#
# class SimpleBeta(CustomFactor, StandardOutputs):
#     """
#     Factor producing the slope of a regression line between each asset's daily
#     returns to the daily returns of a single "target" asset.
#
#     Parameters
#     ----------
#     target : zipline.Asset
#         Asset against which other assets should be regressed.
#     regression_length : int
#         Number of days of daily returns to use for the regression.
#     allowed_missing_percentage : float, optional
#         Percentage of returns observations (between 0 and 1) that are allowed
#         to be missing when calculating betas. Assets with more than this
#         percentage of returns observations missing will produce values of
#         NaN. Default behavior is that 25% of inputs can be missing.
#     """
#     window_safe = True
#     dtype = float64_dtype
#     params = ('allowed_missing_count',)
#
#     @expect_types(
#         target=Asset,
#         regression_length=int,
#         allowed_missing_percentage=(int, float),
#         __funcname='SimpleBeta',
#     )
#     @expect_bounded(
#         regression_length=(3, None),
#         allowed_missing_percentage=(0.0, 1.0),
#         __funcname='SimpleBeta',
#     )
#     def __new__(cls,
#                 target,
#                 regression_length,
#                 allowed_missing_percentage=0.25):
#         daily_returns = Returns(
#             window_length=2,
#             mask=(AssetExists() | SingleAsset(asset=target)),
#         )
#         allowed_missing_count = int(
#             allowed_missing_percentage * regression_length
#         )
#         return super(SimpleBeta, cls).__new__(
#             cls,
#             inputs=[daily_returns, daily_returns[target]],
#             window_length=regression_length,
#             allowed_missing_count=allowed_missing_count,
#         )
#
#     def compute(self,
#                 today,
#                 assets,
#                 out,
#                 all_returns,
#                 target_returns,
#                 allowed_missing_count):
#         vectorized_beta(
#             dependents=all_returns,
#             independent=target_returns,
#             allowed_missing=allowed_missing_count,
#             out=out,
#         )
#
#     def graph_repr(self):
#         return "{}({!r}, {}, {})".format(
#             type(self).__name__,
#             str(self.target.symbol),  # coerce from unicode to str in py2.
#             self.window_length,
#             self.params['allowed_missing_count'],
#         )
#
#     @property
#     def target(self):
#         """Get the target of the beta calculation.
#         """
#         return self.inputs[1].asset
#
#     def __repr__(self):
#         return "{}({}, length={}, allowed_missing={})".format(
#             type(self).__name__,
#             self.target,
#             self.window_length,
#             self.params['allowed_missing_count'],
#         )
#
#
# def vectorized_beta(dependents, independent, allowed_missing, out=None):
#     """
#     Compute slopes of linear regressions between columns of ``dependents`` and
#     ``independent``.
#
#     Parameters
#     ----------
#     dependents : np.array[N, M]
#         Array with columns of data to be regressed against ``independent``.
#     independent : np.array[N, 1]
#         Independent variable of the regression
#     allowed_missing : int
#         Number of allowed missing (NaN) observations per column. Columns with
#         more than this many non-nan observations in either ``dependents`` or
#         ``independents`` will output NaN as the regression coefficient.
#     out : np.array[M] or None, optional
#         Output array into which to write results.  If None, a new array is
#         created and returned.
#
#     Returns
#     -------
#     slopes : np.array[M]
#         Linear regression coefficients for each column of ``dependents``.
#     """
#     # Cache these as locals since we're going to call them multiple times.
#     nan = np.nan
#     isnan = np.isnan
#     N, M = dependents.shape
#
#     if out is None:
#         out = np.full(M, nan)
#
#     # Copy N times as a column vector and fill with nans to have the same
#     # missing value pattern as the dependent variable.
#     #
#     # PERF_TODO: We could probably avoid the space blowup by doing this in
#     # Cython.
#
#     # shape: (N, M)
#     independent = np.where(
#         isnan(dependents),
#         nan,
#         independent,
#     )
#
#     # Calculate beta as Cov(X, Y) / Cov(X, X).
#     # https://en.wikipedia.org/wiki/Simple_linear_regression#Fitting_the_regression_line  # noqa
#     #
#     # NOTE: The usual formula for covariance is::
#     #
#     #    mean((X - mean(X)) * (Y - mean(Y)))
#     #
#     # However, we don't actually need to take the mean of both sides of the
#     # product, because of the folllowing equivalence::
#     #
#     # Let X_res = (X - mean(X)).
#     # We have:
#     #
#     #     mean(X_res * (Y - mean(Y))) = mean(X_res * (Y - mean(Y)))
#     #                             (1) = mean((X_res * Y) - (X_res * mean(Y)))
#     #                             (2) = mean(X_res * Y) - mean(X_res * mean(Y))
#     #                             (3) = mean(X_res * Y) - mean(X_res) * mean(Y)
#     #                             (4) = mean(X_res * Y) - 0 * mean(Y)
#     #                             (5) = mean(X_res * Y)
#     #
#     #
#     # The tricky step in the above derivation is step (4). We know that
#     # mean(X_res) is zero because, for any X:
#     #
#     #     mean(X - mean(X)) = mean(X) - mean(X) = 0.
#     #
#     # The upshot of this is that we only have to center one of `independent`
#     # and `dependent` when calculating covariances. Since we need the centered
#     # `independent` to calculate its variance in the next step, we choose to
#     # center `independent`.
#
#     # shape: (N, M)
#     ind_residual = independent - nanmean(independent, axis=0)
#
#     # shape: (M,)
#     covariances = nanmean(ind_residual * dependents, axis=0)
#
#     # We end up with different variances in each column here because each
#     # column may have a different subset of the data dropped due to missing
#     # data in the corresponding dependent column.
#     # shape: (M,)
#     independent_variances = nanmean(ind_residual ** 2, axis=0)
#
#     # shape: (M,)
#     np.divide(covariances, independent_variances, out=out)
#
#     # Write nans back to locations where we have more then allowed number of
#     # missing entries.
#     nanlocs = isnan(independent).sum(axis=0) > allowed_missing
#     out[nanlocs] = nan
#
#     return out
#
#
# def vectorized_pearson_r(dependents, independents, allowed_missing, out=None):
#     """
#     Compute Pearson's r between columns of ``dependents`` and ``independents``.
#
#     Parameters
#     ----------
#     dependents : np.array[N, M]
#         Array with columns of data to be regressed against ``independent``.
#     independents : np.array[N, M] or np.array[N, 1]
#         Independent variable(s) of the regression. If a single column is
#         passed, it is broadcast to the shape of ``dependents``.
#     allowed_missing : int
#         Number of allowed missing (NaN) observations per column. Columns with
#         more than this many non-nan observations in either ``dependents`` or
#         ``independents`` will output NaN as the correlation coefficient.
#     out : np.array[M] or None, optional
#         Output array into which to write results.  If None, a new array is
#         created and returned.
#
#     Returns
#     -------
#     correlations : np.array[M]
#         Pearson correlation coefficients for each column of ``dependents``.
#
#     See Also
#     --------
#     :class:`zipline.pipeline.factors.RollingPearson`
#     :class:`zipline.pipeline.factors.RollingPearsonOfReturns`
#     """
#     nan = np.nan
#     isnan = np.isnan
#     N, M = dependents.shape
#
#     if out is None:
#         out = np.full(M, nan)
#
#     if allowed_missing > 0:
#         # If we're handling nans robustly, we need to mask both arrays to
#         # locations where either was nan.
#         either_nan = isnan(dependents) | isnan(independents)
#         independents = np.where(either_nan, nan, independents)
#         dependents = np.where(either_nan, nan, dependents)
#         mean = nanmean
#     else:
#         # Otherwise, we can just use mean, which will give us a nan for any
#         # column where there's ever a nan.
#         mean = np.mean
#
#     # Pearson R is Cov(X, Y) / StdDev(X) * StdDev(Y)
#     # c.f. https://en.wikipedia.org/wiki/Pearson_correlation_coefficient
#     ind_residual = independents - mean(independents, axis=0)
#     dep_residual = dependents - mean(dependents, axis=0)
#
#     ind_variance = mean(ind_residual ** 2, axis=0)
#     dep_variance = mean(dep_residual ** 2, axis=0)
#
#     covariances = mean(ind_residual * dep_residual, axis=0)
#
#     evaluate(
#         'where(mask, nan, cov / sqrt(ind_variance * dep_variance))',
#         local_dict={'cov': covariances,
#                     'mask': isnan(independents).sum(axis=0) > allowed_missing,
#                     'nan': np.nan,
#                     'ind_variance': ind_variance,
#                     'dep_variance': dep_variance},
#         global_dict={},
#         out=out,
#     )
#     return out



"""Genetic Programming in Python, with a scikit-learn inspired API

The :mod:`gplearn.genetic` module implements Genetic Programming. These
are supervised learning methods based on applying evolutionary operations on
computer programs.
"""

# Author: Trevor Stephens <trevorstephens.com>
#
# License: BSD 3 clause

import itertools
from abc import ABCMeta, abstractmethod
from time import time
from warnings import warn

import numpy as np
from joblib import Parallel, delayed
from scipy.stats import rankdata
from sklearn.base import BaseEstimator
from sklearn.base import RegressorMixin, TransformerMixin, ClassifierMixin
from sklearn.exceptions import NotFittedError
from sklearn.utils import compute_sample_weight
from sklearn.utils.validation import check_array, _check_sample_weight
from sklearn.utils.multiclass import check_classification_targets

from ._program import _Program
from .fitness import _fitness_map, _Fitness
from .functions import _function_map, _Function, sig1 as sigmoid
from .utils import _partition_estimators
from .utils import check_random_state
import copy
from torchqtm.tdbt.backtest import BaseTester

__all__ = ['SymbolicRegressor', 'SymbolicClassifier', 'SymbolicTransformer']

MAX_INT = np.iinfo(np.int32).max


def _parallel_evolve(n_programs, parents, X, y, sample_weight, seeds, params):
    """Private function used to build a batch of programs within a job."""
    n_samples, n_features = X.shape[0], X.shape[-1]
    # Unpack parameters
    tournament_size = params['tournament_size']
    function_set = params['function_set']
    arities = params['arities']
    init_depth = params['init_depth']
    init_method = params['init_method']
    const_range = params['const_range']
    metric = params['_metric']
    transformer = params['_transformer']
    parsimony_coefficient = params['parsimony_coefficient']
    method_probs = params['method_probs']
    p_point_replace = params['p_point_replace']
    max_samples = params['max_samples']
    feature_names = params['feature_names']
    strategy: BaseTester = params['strategy']
    max_samples = int(max_samples * n_samples)

    def _tournament():
        """Find the fittest individual from a sub-population."""
        contenders = random_state.randint(0, len(parents), tournament_size)
        fitness = [parents[p].fitness_ for p in contenders]
        if metric.greater_is_better:
            parent_index = contenders[np.argmax(fitness)]
        else:
            parent_index = contenders[np.argmin(fitness)]
        return parents[parent_index], parent_index

    # Build programs
    programs = []

    for i in range(n_programs):

        random_state = check_random_state(seeds[i])

        if parents is None:
            program = None
            genome = None
        else:
            method = random_state.uniform()
            parent, parent_index = _tournament()

            if method < method_probs[0]:
                # crossover
                donor, donor_index = _tournament()
                program, removed, remains = parent.crossover(donor.program,
                                                             random_state)
                genome = {'method': 'Crossover',
                          'parent_idx': parent_index,
                          'parent_nodes': removed,
                          'donor_idx': donor_index,
                          'donor_nodes': remains}
            elif method < method_probs[1]:
                # subtree_mutation
                program, removed, _ = parent.subtree_mutation(random_state)
                genome = {'method': 'Subtree Mutation',
                          'parent_idx': parent_index,
                          'parent_nodes': removed}
            elif method < method_probs[2]:
                # hoist_mutation
                program, removed = parent.hoist_mutation(random_state)
                genome = {'method': 'Hoist Mutation',
                          'parent_idx': parent_index,
                          'parent_nodes': removed}
            elif method < method_probs[3]:
                # point_mutation
                program, mutated = parent.point_mutation(random_state)
                genome = {'method': 'Point Mutation',
                          'parent_idx': parent_index,
                          'parent_nodes': mutated}
            else:
                # reproduction
                program = parent.reproduce()
                genome = {'method': 'Reproduction',
                          'parent_idx': parent_index,
                          'parent_nodes': []}

        program = _Program(function_set=function_set,
                           arities=arities,
                           init_depth=init_depth,
                           init_method=init_method,
                           n_features=n_features,
                           metric=metric,
                           transformer=transformer,
                           const_range=const_range,
                           p_point_replace=p_point_replace,
                           parsimony_coefficient=parsimony_coefficient,
                           feature_names=feature_names,
                           random_state=random_state,
                           program=program,
                           strategy=strategy)

        program.parents = genome

        # Draw samples, using sample weights, and then fit
        if sample_weight is None:
            curr_sample_weight = np.ones((n_samples,))
        else:
            curr_sample_weight = sample_weight.copy()
        oob_sample_weight = curr_sample_weight.copy()

        indices, not_indices = program.get_all_indices(n_samples,
                                                       max_samples,
                                                       random_state)

        curr_sample_weight[not_indices] = 0
        oob_sample_weight[indices] = 0

        program.raw_fitness_ = program.raw_fitness(X, y, curr_sample_weight)
        if max_samples < n_samples:
            # Calculate OOB fitness
            program.oob_fitness_ = program.raw_fitness(X, y, oob_sample_weight)

        programs.append(program)

    return programs


class BaseSymbolic(BaseEstimator, metaclass=ABCMeta):
    """Base class for symbolic regression / classification estimators.

    Warning: This class should not be used directly.
    Use derived classes instead.

    """

    @abstractmethod
    def __init__(self,
                 *,
                 population_size=1000,
                 hall_of_fame=None,
                 n_components=None,
                 generations=20,
                 tournament_size=20,
                 stopping_criteria=0.0,
                 const_range=(-1., 1.),
                 init_depth=(2, 6),
                 init_method='half and half',
                 function_set=('add', 'sub', 'mul', 'div'),
                 transformer=None,
                 metric='mean absolute error',
                 parsimony_coefficient=0.001,
                 p_crossover=0.9,
                 p_subtree_mutation=0.01,
                 p_hoist_mutation=0.01,
                 p_point_mutation=0.01,
                 p_point_replace=0.05,
                 max_samples=1.0,
                 class_weight=None,
                 feature_names=None,
                 warm_start=False,
                 low_memory=False,
                 n_jobs=1,
                 verbose=0,
                 random_state=None,
                 strategy: BaseTester = None,
                 ):
        self.population_size = population_size
        self.hall_of_fame = hall_of_fame
        self.n_components = n_components
        self.generations = generations
        self.tournament_size = tournament_size
        self.stopping_criteria = stopping_criteria
        self.const_range = const_range
        self.init_depth = init_depth
        self.init_method = init_method
        self.function_set = function_set
        self.transformer = transformer
        self.metric = metric
        self.parsimony_coefficient = parsimony_coefficient
        self.p_crossover = p_crossover
        self.p_subtree_mutation = p_subtree_mutation
        self.p_hoist_mutation = p_hoist_mutation
        self.p_point_mutation = p_point_mutation
        self.p_point_replace = p_point_replace
        self.max_samples = max_samples
        self.class_weight = class_weight
        self.feature_names = feature_names
        self.warm_start = warm_start
        self.low_memory = low_memory
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.random_state = random_state
        self.strategy: BaseTester = strategy
        self.n_features_in_ = ...
        self._function_set = ...
        self._arities = ...

    def _verbose_reporter(self, run_details=None):
        """A report of the progress of the evolution process.

        Parameters
        ----------
        run_details : dict
            Information about the evolution.

        """
        if run_details is None:
            print('    |{:^25}|{:^42}|'.format('Population Average',
                                               'Best Individual'))
            print('-' * 4 + ' ' + '-' * 25 + ' ' + '-' * 42 + ' ' + '-' * 10)
            line_format = '{:>4} {:>8} {:>16} {:>8} {:>16} {:>16} {:>10}'
            print(line_format.format('Gen', 'Length', 'Fitness', 'Length',
                                     'Fitness', 'OOB Fitness', 'Time Left'))

        else:
            # Estimate remaining time for run_
            gen = run_details['generation'][-1]
            generation_time = run_details['generation_time'][-1]
            remaining_time = (self.generations - gen - 1) * generation_time
            if remaining_time > 60:
                remaining_time = '{0:.2f}m'.format(remaining_time / 60.0)
            else:
                remaining_time = '{0:.2f}s'.format(remaining_time)

            oob_fitness = 'N/A'
            line_format = '{:4d} {:8.2f} {:16g} {:8d} {:16g} {:>16} {:>10}'
            if self.max_samples < 1.0:
                oob_fitness = run_details['best_oob_fitness'][-1]
                line_format = '{:4d} {:8.2f} {:16g} {:8d} {:16g} {:16g} {:>10}'

            print(line_format.format(run_details['generation'][-1],
                                     run_details['average_length'][-1],
                                     run_details['average_fitness'][-1],
                                     run_details['best_length'][-1],
                                     run_details['best_fitness'][-1],
                                     oob_fitness,
                                     remaining_time))

    def fit(self, X, y, sample_weight=None):
        """Fit the Genetic Program according to X, y.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples]
            Target values.

        sample_weight : array-like, shape = [n_samples], optional
            Weights applied to individual samples.

        Returns
        -------
        self : object
            Returns self.

        """
        assert isinstance(X, np.ndarray)
        assert isinstance(y, np.ndarray)

        random_state = check_random_state(self.random_state)

        # Check arrays
        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X)
        self.n_features_in_ = X.shape[-1]
        # if isinstance(self, ClassifierMixin):
        #     X, y = self._validate_data(X, y, y_numeric=False)
        #     check_classification_targets(y)
        #
        #     if self.class_weight:
        #         if sample_weight is None:
        #             sample_weight = 1.
        #         # modify the sample weights with the corresponding class weight
        #         sample_weight = (sample_weight *
        #                          compute_sample_weight(self.class_weight, y))
        #
        #     self.classes_, y = np.unique(y, return_inverse=True)
        #     n_trim_classes = np.count_nonzero(np.bincount(y, sample_weight))
        #     if n_trim_classes != 2:
        #         raise ValueError("y contains %d class after sample_weight "
        #                          "trimmed classes with zero weights, while 2 "
        #                          "classes are required."
        #                          % n_trim_classes)
        #     self.n_classes_ = len(self.classes_)
        #
        # else:
        #     X, y = self._validate_data(X, y, y_numeric=True)

        hall_of_fame = self.hall_of_fame
        if hall_of_fame is None:
            hall_of_fame = self.population_size
        if hall_of_fame > self.population_size or hall_of_fame < 1:
            raise ValueError('hall_of_fame (%d) must be less than or equal to '
                             'population_size (%d).' % (self.hall_of_fame,
                                                        self.population_size))
        n_components = self.n_components
        if n_components is None:
            n_components = hall_of_fame
        if n_components > hall_of_fame or n_components < 1:
            raise ValueError('n_components (%d) must be less than or equal to '
                             'hall_of_fame (%d).' % (self.n_components,
                                                     self.hall_of_fame))

        self._function_set = []
        for function in self.function_set:
            if isinstance(function, str):
                if function not in _function_map:
                    raise ValueError('invalid function name %s found in '
                                     '`function_set`.' % function)
                self._function_set.append(_function_map[function])
            elif isinstance(function, _Function):
                self._function_set.append(function)
            else:
                raise ValueError('invalid type %s found in `function_set`.'
                                 % type(function))
        if not self._function_set:
            raise ValueError('No valid functions found in `function_set`.')

        # For point-mutation to find a compatible replacement node
        self._arities = {}
        for function in self._function_set:
            arity = function.arity
            self._arities[arity] = self._arities.get(arity, [])
            self._arities[arity].append(function)

        if isinstance(self.metric, _Fitness):
            self._metric = self.metric
        elif isinstance(self, RegressorMixin):
            if self.metric not in ('mean absolute error', 'mse', 'rmse',
                                   'pearson', 'spearman'):
                raise ValueError('Unsupported metric: %s' % self.metric)
            self._metric = _fitness_map[self.metric]
        elif isinstance(self, ClassifierMixin):
            if self.metric != 'log loss':
                raise ValueError('Unsupported metric: %s' % self.metric)
            self._metric = _fitness_map[self.metric]
        elif isinstance(self, TransformerMixin):
            if self.metric not in ('pearson', 'spearman'):
                raise ValueError('Unsupported metric: %s' % self.metric)
            self._metric = _fitness_map[self.metric]

        self._method_probs = np.array([self.p_crossover,
                                       self.p_subtree_mutation,
                                       self.p_hoist_mutation,
                                       self.p_point_mutation])
        self._method_probs = np.cumsum(self._method_probs)

        if self._method_probs[-1] > 1:
            raise ValueError('The sum of p_crossover, p_subtree_mutation, '
                             'p_hoist_mutation and p_point_mutation should '
                             'total to 1.0 or less.')

        if self.init_method not in ('half and half', 'grow', 'full'):
            raise ValueError('Valid program initializations methods include '
                             '"grow", "full" and "half and half". Given %s.'
                             % self.init_method)

        if not ((isinstance(self.const_range, tuple) and
                 len(self.const_range) == 2) or self.const_range is None):
            raise ValueError('const_range should be a tuple with length two, '
                             'or None.')

        if (not isinstance(self.init_depth, tuple) or
                len(self.init_depth) != 2):
            raise ValueError('init_depth should be a tuple with length two.')
        if self.init_depth[0] > self.init_depth[1]:
            raise ValueError('init_depth should be in increasing numerical '
                             'order: (min_depth, max_depth).')

        if self.feature_names is not None:
            if self.n_features_in_ != len(self.feature_names):
                raise ValueError('The supplied `feature_names` has different '
                                 'length to n_features. Expected %d, got %d.'
                                 % (self.n_features_in_,
                                    len(self.feature_names)))
            for feature_name in self.feature_names:
                if not isinstance(feature_name, str):
                    raise ValueError('invalid type %s found in '
                                     '`feature_names`.' % type(feature_name))

        if self.transformer is not None:
            if isinstance(self.transformer, _Function):
                self._transformer = self.transformer
            elif self.transformer == 'sigmoid':
                self._transformer = sigmoid
            else:
                raise ValueError('Invalid `transformer`. Expected either '
                                 '"sigmoid" or _Function object, got %s' %
                                 type(self.transformer))
            if self._transformer.arity != 1:
                raise ValueError('Invalid arity for `transformer`. Expected 1, '
                                 'got %d.' % (self._transformer.arity))

        params = self.get_params()
        params['_metric'] = self._metric
        if hasattr(self, '_transformer'):
            params['_transformer'] = self._transformer
        else:
            params['_transformer'] = None
        params['function_set'] = self._function_set
        params['arities'] = self._arities
        params['method_probs'] = self._method_probs

        if not self.warm_start or not hasattr(self, '_programs'):
            # Free allocated memory, if any
            self._programs = []
            self.run_details_ = {'generation': [],
                                 'average_length': [],
                                 'average_fitness': [],
                                 'best_length': [],
                                 'best_fitness': [],
                                 'best_oob_fitness': [],
                                 'generation_time': []}

        prior_generations = len(self._programs)
        n_more_generations = self.generations - prior_generations

        if n_more_generations < 0:
            raise ValueError('generations=%d must be larger or equal to '
                             'len(_programs)=%d when warm_start==True'
                             % (self.generations, len(self._programs)))
        elif n_more_generations == 0:
            fitness = [program.raw_fitness_ for program in self._programs[-1]]
            warn('Warm-start fitting without increasing n_estimators does not '
                 'fit new programs.')

        if self.warm_start:
            # Generate and discard seeds that would have been produced on the
            # initial fit call.
            for i in range(len(self._programs)):
                _ = random_state.randint(MAX_INT, size=self.population_size)

        if self.verbose:
            # Print header fields
            self._verbose_reporter()

        for gen in range(prior_generations, self.generations):

            start_time = time()

            if gen == 0:
                parents = None
            else:
                parents = self._programs[gen - 1]

            # Parallel loop
            n_jobs, n_programs, starts = _partition_estimators(
                self.population_size, self.n_jobs)
            seeds = random_state.randint(MAX_INT, size=self.population_size)

            population = Parallel(n_jobs=n_jobs,
                                  verbose=int(self.verbose > 1))(
                delayed(_parallel_evolve)(n_programs[i],
                                          parents,
                                          X,
                                          y,
                                          sample_weight,
                                          seeds[starts[i]:starts[i + 1]],
                                          params)
                for i in range(n_jobs))

            # Reduce, maintaining order across different n_jobs
            population = list(itertools.chain.from_iterable(population))

            fitness = [program.raw_fitness_ for program in population]
            length = [program.length_ for program in population]

            parsimony_coefficient = None
            if self.parsimony_coefficient == 'auto':
                parsimony_coefficient = (np.cov(length, fitness)[1, 0] /
                                         np.var(length))
            for program in population:
                program.fitness_ = program.fitness(parsimony_coefficient)

            self._programs.append(population)

            # Remove old programs that didn't make it into the new population.
            if not self.low_memory:
                for old_gen in np.arange(gen, 0, -1):
                    indices = []
                    for program in self._programs[old_gen]:
                        if program is not None:
                            for idx in program.parents:
                                if 'idx' in idx:
                                    indices.append(program.parents[idx])
                    indices = set(indices)
                    for idx in range(self.population_size):
                        if idx not in indices:
                            self._programs[old_gen - 1][idx] = None
            elif gen > 0:
                # Remove old generations
                self._programs[gen - 1] = None

            # Record run_ details
            if self._metric.greater_is_better:
                best_program = population[np.argmax(fitness)]
            else:
                best_program = population[np.argmin(fitness)]

            self.run_details_['generation'].append(gen)
            self.run_details_['average_length'].append(np.mean(length))
            self.run_details_['average_fitness'].append(np.mean(fitness))
            self.run_details_['best_length'].append(best_program.length_)
            self.run_details_['best_fitness'].append(best_program.raw_fitness_)
            oob_fitness = np.nan
            if self.max_samples < 1.0:
                oob_fitness = best_program.oob_fitness_
            self.run_details_['best_oob_fitness'].append(oob_fitness)
            generation_time = time() - start_time
            self.run_details_['generation_time'].append(generation_time)

            if self.verbose:
                self._verbose_reporter(self.run_details_)

            # Check for early stopping
            if self._metric.greater_is_better:
                best_fitness = fitness[np.argmax(fitness)]
                if best_fitness >= self.stopping_criteria:
                    break
            else:
                best_fitness = fitness[np.argmin(fitness)]
                if best_fitness <= self.stopping_criteria:
                    break

        if isinstance(self, TransformerMixin):
            # Find the best individuals in the final generation
            fitness = np.array(fitness)
            if self._metric.greater_is_better:
                hall_of_fame = fitness.argsort()[::-1][:self.hall_of_fame]
            else:
                hall_of_fame = fitness.argsort()[:self.hall_of_fame]
            evaluation = np.array([gp.execute(X) for gp in
                                   [self._programs[-1][i] for
                                    i in hall_of_fame]])
            if self.metric == 'spearman':
                evaluation = np.apply_along_axis(rankdata, 1, evaluation)

            with np.errstate(divide='ignore', invalid='ignore'):
                correlations = np.abs(np.corrcoef(evaluation))
            np.fill_diagonal(correlations, 0.)
            components = list(range(self.hall_of_fame))
            indices = list(range(self.hall_of_fame))
            # Iteratively remove least fit individual of most correlated pair
            while len(components) > self.n_components:
                most_correlated = np.unravel_index(np.argmax(correlations),
                                                   correlations.shape)
                # The correlation matrix is sorted by fitness, so identifying
                # the least fit of the pair is simply getting the higher index
                worst = max(most_correlated)
                components.pop(worst)
                indices.remove(worst)
                correlations = correlations[:, indices][indices, :]
                indices = list(range(len(components)))
            self._best_programs = [self._programs[-1][i] for i in
                                   hall_of_fame[components]]

        else:
            # Find the best individual in the final generation
            if self._metric.greater_is_better:
                self._program = self._programs[-1][np.argmax(fitness)]
            else:
                self._program = self._programs[-1][np.argmin(fitness)]

        return self


class SymbolicRegressor(BaseSymbolic, RegressorMixin):
    """A Genetic Programming symbolic regressor.

    A symbolic regressor is an estimator that begins by building a population
    of naive random formulas to represent a relationship. The formulas are
    represented as tree-like structures with mathematical functions being
    recursively applied to variables and constants. Each successive generation
    of programs is then evolved from the one that came before it by selecting
    the fittest individuals from the population to undergo genetic operations
    such as crossover, mutation or reproduction.

    Parameters
    ----------
    population_size : integer, optional (default=1000)
        The number of programs in each generation.

    generations : integer, optional (default=20)
        The number of generations to evolve.

    tournament_size : integer, optional (default=20)
        The number of programs that will compete to become part of the next
        generation.

    stopping_criteria : float, optional (default=0.0)
        The required metric value required in order to stop evolution early.

    const_range : tuple of two floats, or None, optional (default=(-1., 1.))
        The range of constants to include in the formulas. If None then no
        constants will be included in the candidate programs.

    init_depth : tuple of two ints, optional (default=(2, 6))
        The range of tree depths for the initial population of naive formulas.
        Individual trees will randomly choose a maximum depth from this range.
        When combined with `init_method='half and half'` this yields the well-
        known 'ramped half and half' initialization method.

    init_method : str, optional (default='half and half')
        - 'grow' : Nodes are chosen at random from both functions and
          terminals, allowing for smaller trees than `init_depth` allows. Tends
          to grow asymmetrical trees.
        - 'full' : Functions are chosen until the `init_depth` is reached, and
          then terminals are selected. Tends to grow 'bushy' trees.
        - 'half and half' : Trees are grown through a 50/50 mix of 'full' and
          'grow', making for a mix of tree shapes in the initial population.

    function_set : iterable, optional (default=('add', 'sub', 'mul', 'div'))
        The functions to use when building and evolving programs. This iterable
        can include strings to indicate either individual functions as outlined
        below, or you can also include your own functions as built using the
        ``make_function`` factory from the ``functions`` module.

        Available individual functions are:

        - 'add' : addition, arity=2.
        - 'sub' : subtraction, arity=2.
        - 'mul' : multiplication, arity=2.
        - 'div' : protected division where a denominator near-zero returns 1.,
          arity=2.
        - 'sqrt' : protected square root where the absolute value of the
          argument is used, arity=1.
        - 'log' : protected log where the absolute value of the argument is
          used and a near-zero argument returns 0., arity=1.
        - 'abs' : absolute value, arity=1.
        - 'neg' : negative, arity=1.
        - 'inv' : protected inverse where a near-zero argument returns 0.,
          arity=1.
        - 'max' : maximum, arity=2.
        - 'min' : minimum, arity=2.
        - 'sin' : sine (radians), arity=1.
        - 'cos' : cosine (radians), arity=1.
        - 'tan' : tangent (radians), arity=1.

    metric : str, optional (default='mean absolute error')
        The name of the raw fitness metric. Available options include:

        - 'mean absolute error'.
        - 'mse' for mean squared error.
        - 'rmse' for root mean squared error.
        - 'pearson', for Pearson's product-moment correlation coefficient.
        - 'spearman' for Spearman's rank-order correlation coefficient.

        Note that 'pearson' and 'spearman' will not directly predict the target
        but could be useful as value-added features in a second-step estimator.
        This would allow the user to generate one engineered feature at a time,
        using the SymbolicTransformer would allow creation of multiple features
        at once.

    parsimony_coefficient : float or "auto", optional (default=0.001)
        This constant penalizes large programs by adjusting their fitness to
        be less favorable for selection. Larger values penalize the program
        more which can control the phenomenon known as 'bloat'. Bloat is when
        evolution is increasing the size of programs without a significant
        increase in fitness, which is costly for computation time and makes for
        a less understandable final result. This parameter may need to be tuned
        over successive runs.

        If "auto" the parsimony coefficient is recalculated for each generation
        using c = Cov(l,f)/Var( l), where Cov(l,f) is the covariance between
        program size l and program fitness f in the population, and Var(l) is
        the variance of program sizes.

    p_crossover : float, optional (default=0.9)
        The probability of performing crossover on a tournament winner.
        Crossover takes the winner of a tournament and selects a random subtree
        from it to be replaced. A second tournament is performed to find a
        donor. The donor also has a subtree selected at random and this is
        inserted into the original parent to form an offspring in the next
        generation.

    p_subtree_mutation : float, optional (default=0.01)
        The probability of performing subtree mutation on a tournament winner.
        Subtree mutation takes the winner of a tournament and selects a random
        subtree from it to be replaced. A donor subtree is generated at random
        and this is inserted into the original parent to form an offspring in
        the next generation.

    p_hoist_mutation : float, optional (default=0.01)
        The probability of performing hoist mutation on a tournament winner.
        Hoist mutation takes the winner of a tournament and selects a random
        subtree from it. A random subtree of that subtree is then selected
        and this is 'hoisted' into the original subtrees location to form an
        offspring in the next generation. This method helps to control bloat.

    p_point_mutation : float, optional (default=0.01)
        The probability of performing point mutation on a tournament winner.
        Point mutation takes the winner of a tournament and selects random
        nodes from it to be replaced. Terminals are replaced by other terminals
        and functions are replaced by other functions that require the same
        number of arguments as the original node. The resulting tree forms an
        offspring in the next generation.

        Note : The above genetic operation probabilities must sum to less than
        one. The balance of probability is assigned to 'reproduction', where a
        tournament winner is cloned and enters the next generation unmodified.

    p_point_replace : float, optional (default=0.05)
        For point mutation only, the probability that any given node will be
        mutated.

    max_samples : float, optional (default=1.0)
        The fraction of samples to draw from X to evaluate each program on.

    feature_names : list, optional (default=None)
        Optional list of feature names, used purely for representations in
        the `print` operation or `export_graphviz`. If None, then X0, X1, etc
        will be used for representations.

    warm_start : bool, optional (default=False)
        When set to ``True``, reuse the solution of the previous call to fit
        and add more generations to the evolution, otherwise, just fit a new
        evolution.

    low_memory : bool, optional (default=False)
        When set to ``True``, only the current generation is retained. Parent
        information is discarded. For very large populations or runs with many
        generations, this can result in substantial memory use reduction.

    n_jobs : integer, optional (default=1)
        The number of jobs to run_ in parallel for `fit`. If -1, then the number
        of jobs is set to the number of cores.

    verbose : int, optional (default=0)
        Controls the verbosity of the evolution building process.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Attributes
    ----------
    run_details_ : dict
        Details of the evolution process. Includes the following elements:

        - 'generation' : The generation index.
        - 'average_length' : The average program length of the generation.
        - 'average_fitness' : The average program fitness of the generation.
        - 'best_length' : The length of the best program in the generation.
        - 'best_fitness' : The fitness of the best program in the generation.
        - 'best_oob_fitness' : The out of bag fitness of the best program in
          the generation (requires `max_samples` < 1.0).
        - 'generation_time' : The time it took for the generation to evolve.

    See Also
    --------
    SymbolicTransformer

    References
    ----------
    .. [1] J. Koza, "Genetic Programming", 1992.

    .. [2] R. Poli, et al. "A Field Guide to Genetic Programming", 2008.

    """

    def __init__(self,
                 *,
                 population_size=1000,
                 generations=20,
                 tournament_size=20,
                 stopping_criteria=0.0,
                 const_range=(-1., 1.),
                 init_depth=(2, 6),
                 init_method='half and half',
                 function_set=('add', 'sub', 'mul', 'div'),
                 metric='mean absolute error',
                 parsimony_coefficient=0.001,
                 p_crossover=0.9,
                 p_subtree_mutation=0.01,
                 p_hoist_mutation=0.01,
                 p_point_mutation=0.01,
                 p_point_replace=0.05,
                 max_samples=1.0,
                 feature_names=None,
                 warm_start=False,
                 low_memory=False,
                 n_jobs=1,
                 verbose=0,
                 random_state=None,
                 strategy: BaseTester = None):
        super(SymbolicRegressor, self).__init__(
            population_size=population_size,
            generations=generations,
            tournament_size=tournament_size,
            stopping_criteria=stopping_criteria,
            const_range=const_range,
            init_depth=init_depth,
            init_method=init_method,
            function_set=function_set,
            metric=metric,
            parsimony_coefficient=parsimony_coefficient,
            p_crossover=p_crossover,
            p_subtree_mutation=p_subtree_mutation,
            p_hoist_mutation=p_hoist_mutation,
            p_point_mutation=p_point_mutation,
            p_point_replace=p_point_replace,
            max_samples=max_samples,
            feature_names=feature_names,
            warm_start=warm_start,
            low_memory=low_memory,
            n_jobs=n_jobs,
            verbose=verbose,
            random_state=random_state,
            strategy=strategy)

    def __str__(self):
        """Overloads `print` output of the object to resemble a LISP tree."""
        if not hasattr(self, '_program'):
            return self.__repr__()
        return self._program.__str__()

    def predict(self, X):
        """Perform regression on mytest.ipynb vectors X.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Input vectors, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        y : array, shape = [n_samples]
            Predicted values for X.

        """
        if not hasattr(self, '_program'):
            raise NotFittedError('SymbolicRegressor not fitted.')

        X = check_array(X)
        _, n_features = X.shape
        if self.n_features_in_ != n_features:
            raise ValueError('Number of features of the model must match the '
                             'input. Model n_features is %s and input '
                             'n_features is %s.'
                             % (self.n_features_in_, n_features))

        y = self._program.execute(X)

        return y


class SymbolicClassifier(BaseSymbolic, ClassifierMixin):
    """A Genetic Programming symbolic classifier.

    A symbolic classifier is an estimator that begins by building a population
    of naive random formulas to represent a relationship. The formulas are
    represented as tree-like structures with mathematical functions being
    recursively applied to variables and constants. Each successive generation
    of programs is then evolved from the one that came before it by selecting
    the fittest individuals from the population to undergo genetic operations
    such as crossover, mutation or reproduction.

    Parameters
    ----------
    population_size : integer, optional (default=500)
        The number of programs in each generation.

    generations : integer, optional (default=10)
        The number of generations to evolve.

    tournament_size : integer, optional (default=20)
        The number of programs that will compete to become part of the next
        generation.

    stopping_criteria : float, optional (default=0.0)
        The required metric value required in order to stop evolution early.

    const_range : tuple of two floats, or None, optional (default=(-1., 1.))
        The range of constants to include in the formulas. If None then no
        constants will be included in the candidate programs.

    init_depth : tuple of two ints, optional (default=(2, 6))
        The range of tree depths for the initial population of naive formulas.
        Individual trees will randomly choose a maximum depth from this range.
        When combined with `init_method='half and half'` this yields the well-
        known 'ramped half and half' initialization method.

    init_method : str, optional (default='half and half')
        - 'grow' : Nodes are chosen at random from both functions and
          terminals, allowing for smaller trees than `init_depth` allows. Tends
          to grow asymmetrical trees.
        - 'full' : Functions are chosen until the `init_depth` is reached, and
          then terminals are selected. Tends to grow 'bushy' trees.
        - 'half and half' : Trees are grown through a 50/50 mix of 'full' and
          'grow', making for a mix of tree shapes in the initial population.

    function_set : iterable, optional (default=('add', 'sub', 'mul', 'div'))
        The functions to use when building and evolving programs. This iterable
        can include strings to indicate either individual functions as outlined
        below, or you can also include your own functions as built using the
        ``make_function`` factory from the ``functions`` module.

        Available individual functions are:

        - 'add' : addition, arity=2.
        - 'sub' : subtraction, arity=2.
        - 'mul' : multiplication, arity=2.
        - 'div' : protected division where a denominator near-zero returns 1.,
          arity=2.
        - 'sqrt' : protected square root where the absolute value of the
          argument is used, arity=1.
        - 'log' : protected log where the absolute value of the argument is
          used and a near-zero argument returns 0., arity=1.
        - 'abs' : absolute value, arity=1.
        - 'neg' : negative, arity=1.
        - 'inv' : protected inverse where a near-zero argument returns 0.,
          arity=1.
        - 'max' : maximum, arity=2.
        - 'min' : minimum, arity=2.
        - 'sin' : sine (radians), arity=1.
        - 'cos' : cosine (radians), arity=1.
        - 'tan' : tangent (radians), arity=1.

    transformer : str, optional (default='sigmoid')
        The name of the function through which the raw decision function is
        passed. This function will transform the raw decision function into
        probabilities of each class.

        This can also be replaced by your own functions as built using the
        ``make_function`` factory from the ``functions`` module.

    metric : str, optional (default='log loss')
        The name of the raw fitness metric. Available options include:

        - 'log loss' aka binary cross-entropy loss.

    parsimony_coefficient : float or "auto", optional (default=0.001)
        This constant penalizes large programs by adjusting their fitness to
        be less favorable for selection. Larger values penalize the program
        more which can control the phenomenon known as 'bloat'. Bloat is when
        evolution is increasing the size of programs without a significant
        increase in fitness, which is costly for computation time and makes for
        a less understandable final result. This parameter may need to be tuned
        over successive runs.

        If "auto" the parsimony coefficient is recalculated for each generation
        using c = Cov(l,f)/Var( l), where Cov(l,f) is the covariance between
        program size l and program fitness f in the population, and Var(l) is
        the variance of program sizes.

    p_crossover : float, optional (default=0.9)
        The probability of performing crossover on a tournament winner.
        Crossover takes the winner of a tournament and selects a random subtree
        from it to be replaced. A second tournament is performed to find a
        donor. The donor also has a subtree selected at random and this is
        inserted into the original parent to form an offspring in the next
        generation.

    p_subtree_mutation : float, optional (default=0.01)
        The probability of performing subtree mutation on a tournament winner.
        Subtree mutation takes the winner of a tournament and selects a random
        subtree from it to be replaced. A donor subtree is generated at random
        and this is inserted into the original parent to form an offspring in
        the next generation.

    p_hoist_mutation : float, optional (default=0.01)
        The probability of performing hoist mutation on a tournament winner.
        Hoist mutation takes the winner of a tournament and selects a random
        subtree from it. A random subtree of that subtree is then selected
        and this is 'hoisted' into the original subtrees location to form an
        offspring in the next generation. This method helps to control bloat.

    p_point_mutation : float, optional (default=0.01)
        The probability of performing point mutation on a tournament winner.
        Point mutation takes the winner of a tournament and selects random
        nodes from it to be replaced. Terminals are replaced by other terminals
        and functions are replaced by other functions that require the same
        number of arguments as the original node. The resulting tree forms an
        offspring in the next generation.

        Note : The above genetic operation probabilities must sum to less than
        one. The balance of probability is assigned to 'reproduction', where a
        tournament winner is cloned and enters the next generation unmodified.

    p_point_replace : float, optional (default=0.05)
        For point mutation only, the probability that any given node will be
        mutated.

    max_samples : float, optional (default=1.0)
        The fraction of samples to draw from X to evaluate each program on.

    class_weight : dict, 'balanced' or None, optional (default=None)
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have weight one.

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input rawdata
        as ``n_samples / (n_classes * np.bincount(y))``

    feature_names : list, optional (default=None)
        Optional list of feature names, used purely for representations in
        the `print` operation or `export_graphviz`. If None, then X0, X1, etc
        will be used for representations.

    warm_start : bool, optional (default=False)
        When set to ``True``, reuse the solution of the previous call to fit
        and add more generations to the evolution, otherwise, just fit a new
        evolution.

    low_memory : bool, optional (default=False)
        When set to ``True``, only the current generation is retained. Parent
        information is discarded. For very large populations or runs with many
        generations, this can result in substantial memory use reduction.

    n_jobs : integer, optional (default=1)
        The number of jobs to run_ in parallel for `fit`. If -1, then the number
        of jobs is set to the number of cores.

    verbose : int, optional (default=0)
        Controls the verbosity of the evolution building process.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Attributes
    ----------
    run_details_ : dict
        Details of the evolution process. Includes the following elements:

        - 'generation' : The generation index.
        - 'average_length' : The average program length of the generation.
        - 'average_fitness' : The average program fitness of the generation.
        - 'best_length' : The length of the best program in the generation.
        - 'best_fitness' : The fitness of the best program in the generation.
        - 'best_oob_fitness' : The out of bag fitness of the best program in
          the generation (requires `max_samples` < 1.0).
        - 'generation_time' : The time it took for the generation to evolve.

    See Also
    --------
    SymbolicTransformer

    References
    ----------
    .. [1] J. Koza, "Genetic Programming", 1992.

    .. [2] R. Poli, et al. "A Field Guide to Genetic Programming", 2008.

    """

    def __init__(self,
                 *,
                 population_size=1000,
                 generations=20,
                 tournament_size=20,
                 stopping_criteria=0.0,
                 const_range=(-1., 1.),
                 init_depth=(2, 6),
                 init_method='half and half',
                 function_set=('add', 'sub', 'mul', 'div'),
                 transformer='sigmoid',
                 metric='log loss',
                 parsimony_coefficient=0.001,
                 p_crossover=0.9,
                 p_subtree_mutation=0.01,
                 p_hoist_mutation=0.01,
                 p_point_mutation=0.01,
                 p_point_replace=0.05,
                 max_samples=1.0,
                 class_weight=None,
                 feature_names=None,
                 warm_start=False,
                 low_memory=False,
                 n_jobs=1,
                 verbose=0,
                 random_state=None):
        super(SymbolicClassifier, self).__init__(
            population_size=population_size,
            generations=generations,
            tournament_size=tournament_size,
            stopping_criteria=stopping_criteria,
            const_range=const_range,
            init_depth=init_depth,
            init_method=init_method,
            function_set=function_set,
            transformer=transformer,
            metric=metric,
            parsimony_coefficient=parsimony_coefficient,
            p_crossover=p_crossover,
            p_subtree_mutation=p_subtree_mutation,
            p_hoist_mutation=p_hoist_mutation,
            p_point_mutation=p_point_mutation,
            p_point_replace=p_point_replace,
            max_samples=max_samples,
            class_weight=class_weight,
            feature_names=feature_names,
            warm_start=warm_start,
            low_memory=low_memory,
            n_jobs=n_jobs,
            verbose=verbose,
            random_state=random_state)

    def __str__(self):
        """Overloads `print` output of the object to resemble a LISP tree."""
        if not hasattr(self, '_program'):
            return self.__repr__()
        return self._program.__str__()

    def _more_tags(self):
        return {'binary_only': True}

    def predict_proba(self, X):
        """Predict probabilities on mytest.ipynb vectors X.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Input vectors, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        proba : array, shape = [n_samples, n_classes]
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute `classes_`.

        """
        if not hasattr(self, '_program'):
            raise NotFittedError('SymbolicClassifier not fitted.')

        X = check_array(X)
        _, n_features = X.shape
        if self.n_features_in_ != n_features:
            raise ValueError('Number of features of the model must match the '
                             'input. Model n_features is %s and input '
                             'n_features is %s.'
                             % (self.n_features_in_, n_features))

        scores = self._program.execute(X)
        proba = self._transformer(scores)
        proba = np.vstack([1 - proba, proba]).T
        return proba

    def predict(self, X):
        """Predict classes on mytest.ipynb vectors X.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Input vectors, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        y : array, shape = [n_samples,]
            The predicted classes of the input samples.

        """
        proba = self.predict_proba(X)
        return self.classes_.take(np.argmax(proba, axis=1), axis=0)


class SymbolicTransformer(BaseSymbolic, TransformerMixin):
    """A Genetic Programming symbolic transformer.

    A symbolic transformer is a supervised transformer that begins by building
    a population of naive random formulas to represent a relationship. The
    formulas are represented as tree-like structures with mathematical
    functions being recursively applied to variables and constants. Each
    successive generation of programs is then evolved from the one that came
    before it by selecting the fittest individuals from the population to
    undergo genetic operations such as crossover, mutation or reproduction.
    The final population is searched for the fittest individuals with the least
    correlation to one another.

    Parameters
    ----------
    population_size : integer, optional (default=1000)
        The number of programs in each generation.

    hall_of_fame : integer, or None, optional (default=100)
        The number of fittest programs to compare from when finding the
        least-correlated individuals for the n_components. If `None`, the
        entire final generation will be used.

    n_components : integer, or None, optional (default=10)
        The number of best programs to return after searching the hall_of_fame
        for the least-correlated individuals. If `None`, the entire
        hall_of_fame will be used.

    generations : integer, optional (default=20)
        The number of generations to evolve.

    tournament_size : integer, optional (default=20)
        The number of programs that will compete to become part of the next
        generation.

    stopping_criteria : float, optional (default=1.0)
        The required metric value required in order to stop evolution early.

    const_range : tuple of two floats, or None, optional (default=(-1., 1.))
        The range of constants to include in the formulas. If None then no
        constants will be included in the candidate programs.

    init_depth : tuple of two ints, optional (default=(2, 6))
        The range of tree depths for the initial population of naive formulas.
        Individual trees will randomly choose a maximum depth from this range.
        When combined with `init_method='half and half'` this yields the well-
        known 'ramped half and half' initialization method.

    init_method : str, optional (default='half and half')
        - 'grow' : Nodes are chosen at random from both functions and
          terminals, allowing for smaller trees than `init_depth` allows. Tends
          to grow asymmetrical trees.
        - 'full' : Functions are chosen until the `init_depth` is reached, and
          then terminals are selected. Tends to grow 'bushy' trees.
        - 'half and half' : Trees are grown through a 50/50 mix of 'full' and
          'grow', making for a mix of tree shapes in the initial population.

    function_set : iterable, optional (default=('add', 'sub', 'mul', 'div'))
        The functions to use when building and evolving programs. This iterable
        can include strings to indicate either individual functions as outlined
        below, or you can also include your own functions as built using the
        ``make_function`` factory from the ``functions`` module.

        Available individual functions are:

        - 'add' : addition, arity=2.
        - 'sub' : subtraction, arity=2.
        - 'mul' : multiplication, arity=2.
        - 'div' : protected division where a denominator near-zero returns 1.,
          arity=2.
        - 'sqrt' : protected square root where the absolute value of the
          argument is used, arity=1.
        - 'log' : protected log where the absolute value of the argument is
          used and a near-zero argument returns 0., arity=1.
        - 'abs' : absolute value, arity=1.
        - 'neg' : negative, arity=1.
        - 'inv' : protected inverse where a near-zero argument returns 0.,
          arity=1.
        - 'max' : maximum, arity=2.
        - 'min' : minimum, arity=2.
        - 'sin' : sine (radians), arity=1.
        - 'cos' : cosine (radians), arity=1.
        - 'tan' : tangent (radians), arity=1.

    metric : str, optional (default='pearson')
        The name of the raw fitness metric. Available options include:

        - 'pearson', for Pearson's product-moment correlation coefficient.
        - 'spearman' for Spearman's rank-order correlation coefficient.

    parsimony_coefficient : float or "auto", optional (default=0.001)
        This constant penalizes large programs by adjusting their fitness to
        be less favorable for selection. Larger values penalize the program
        more which can control the phenomenon known as 'bloat'. Bloat is when
        evolution is increasing the size of programs without a significant
        increase in fitness, which is costly for computation time and makes for
        a less understandable final result. This parameter may need to be tuned
        over successive runs.

        If "auto" the parsimony coefficient is recalculated for each generation
        using c = Cov(l,f)/Var( l), where Cov(l,f) is the covariance between
        program size l and program fitness f in the population, and Var(l) is
        the variance of program sizes.

    p_crossover : float, optional (default=0.9)
        The probability of performing crossover on a tournament winner.
        Crossover takes the winner of a tournament and selects a random subtree
        from it to be replaced. A second tournament is performed to find a
        donor. The donor also has a subtree selected at random and this is
        inserted into the original parent to form an offspring in the next
        generation.

    p_subtree_mutation : float, optional (default=0.01)
        The probability of performing subtree mutation on a tournament winner.
        Subtree mutation takes the winner of a tournament and selects a random
        subtree from it to be replaced. A donor subtree is generated at random
        and this is inserted into the original parent to form an offspring in
        the next generation.

    p_hoist_mutation : float, optional (default=0.01)
        The probability of performing hoist mutation on a tournament winner.
        Hoist mutation takes the winner of a tournament and selects a random
        subtree from it. A random subtree of that subtree is then selected
        and this is 'hoisted' into the original subtrees location to form an
        offspring in the next generation. This method helps to control bloat.

    p_point_mutation : float, optional (default=0.01)
        The probability of performing point mutation on a tournament winner.
        Point mutation takes the winner of a tournament and selects random
        nodes from it to be replaced. Terminals are replaced by other terminals
        and functions are replaced by other functions that require the same
        number of arguments as the original node. The resulting tree forms an
        offspring in the next generation.

        Note : The above genetic operation probabilities must sum to less than
        one. The balance of probability is assigned to 'reproduction', where a
        tournament winner is cloned and enters the next generation unmodified.

    p_point_replace : float, optional (default=0.05)
        For point mutation only, the probability that any given node will be
        mutated.

    max_samples : float, optional (default=1.0)
        The fraction of samples to draw from X to evaluate each program on.

    feature_names : list, optional (default=None)
        Optional list of feature names, used purely for representations in
        the `print` operation or `export_graphviz`. If None, then X0, X1, etc
        will be used for representations.

    warm_start : bool, optional (default=False)
        When set to ``True``, reuse the solution of the previous call to fit
        and add more generations to the evolution, otherwise, just fit a new
        evolution.

    low_memory : bool, optional (default=False)
        When set to ``True``, only the current generation is retained. Parent
        information is discarded. For very large populations or runs with many
        generations, this can result in substantial memory use reduction.

    n_jobs : integer, optional (default=1)
        The number of jobs to run_ in parallel for `fit`. If -1, then the number
        of jobs is set to the number of cores.

    verbose : int, optional (default=0)
        Controls the verbosity of the evolution building process.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Attributes
    ----------
    run_details_ : dict
        Details of the evolution process. Includes the following elements:

        - 'generation' : The generation index.
        - 'average_length' : The average program length of the generation.
        - 'average_fitness' : The average program fitness of the generation.
        - 'best_length' : The length of the best program in the generation.
        - 'best_fitness' : The fitness of the best program in the generation.
        - 'best_oob_fitness' : The out of bag fitness of the best program in
          the generation (requires `max_samples` < 1.0).
        - 'generation_time' : The time it took for the generation to evolve.

    See Also
    --------
    SymbolicRegressor

    References
    ----------
    .. [1] J. Koza, "Genetic Programming", 1992.

    .. [2] R. Poli, et al. "A Field Guide to Genetic Programming", 2008.

    """

    def __init__(self,
                 *,
                 population_size=1000,
                 hall_of_fame=100,
                 n_components=10,
                 generations=20,
                 tournament_size=20,
                 stopping_criteria=1.0,
                 const_range=(-1., 1.),
                 init_depth=(2, 6),
                 init_method='half and half',
                 function_set=('add', 'sub', 'mul', 'div'),
                 metric='pearson',
                 parsimony_coefficient=0.001,
                 p_crossover=0.9,
                 p_subtree_mutation=0.01,
                 p_hoist_mutation=0.01,
                 p_point_mutation=0.01,
                 p_point_replace=0.05,
                 max_samples=1.0,
                 feature_names=None,
                 warm_start=False,
                 low_memory=False,
                 n_jobs=1,
                 verbose=0,
                 random_state=None):
        super(SymbolicTransformer, self).__init__(
            population_size=population_size,
            hall_of_fame=hall_of_fame,
            n_components=n_components,
            generations=generations,
            tournament_size=tournament_size,
            stopping_criteria=stopping_criteria,
            const_range=const_range,
            init_depth=init_depth,
            init_method=init_method,
            function_set=function_set,
            metric=metric,
            parsimony_coefficient=parsimony_coefficient,
            p_crossover=p_crossover,
            p_subtree_mutation=p_subtree_mutation,
            p_hoist_mutation=p_hoist_mutation,
            p_point_mutation=p_point_mutation,
            p_point_replace=p_point_replace,
            max_samples=max_samples,
            feature_names=feature_names,
            warm_start=warm_start,
            low_memory=low_memory,
            n_jobs=n_jobs,
            verbose=verbose,
            random_state=random_state)

    def __len__(self):
        """Overloads `len` output to be the number of fitted components."""
        if not hasattr(self, '_best_programs'):
            return 0
        return self.n_components

    def __getitem__(self, item):
        """Return the ith item of the fitted components."""
        if item >= len(self):
            raise IndexError
        return self._best_programs[item]

    def __str__(self):
        """Overloads `print` output of the object to resemble LISP trees."""
        if not hasattr(self, '_best_programs'):
            return self.__repr__()
        output = str([gp.__str__() for gp in self])
        return output.replace("',", ",\n").replace("'", "")

    def _more_tags(self):
        return {
            "_xfail_checks": {
                "check_sample_weights_invariance": (
                    "zero sample_weight is not equivalent to removing samples"
                ),
            }
        }

    def transform(self, X):
        """Transform X according to the fitted transformer.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Input vectors, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        X_new : array-like, shape = [n_samples, n_components]
            Transformed array.

        """
        if not hasattr(self, '_best_programs'):
            raise NotFittedError('SymbolicTransformer not fitted.')

        X = check_array(X)
        _, n_features = X.shape
        if self.n_features_in_ != n_features:
            raise ValueError('Number of features of the model must match the '
                             'input. Model n_features is %s and input '
                             'n_features is %s.'
                             % (self.n_features_in_, n_features))

        X_new = np.array([gp.execute(X) for gp in self._best_programs]).T

        return X_new

    def fit_transform(self, X, y, sample_weight=None):
        """Fit to rawdata, then transform it.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples]
            Target values.

        sample_weight : array-like, shape = [n_samples], optional
            Weights applied to individual samples.

        Returns
        -------
        X_new : array-like, shape = [n_samples, n_components]
            Transformed array.

        """
        return self.fit(X, y, sample_weight).transform(X)



# distutils: language=c++
from typing import Type

import numpy as np
cimport numpy as cnp
import cython
# from Cython.Includes.numpy import ndarray
from libc.math cimport round
from libcpp.deque cimport deque
from numpy cimport (
    float32_t,
    float64_t,
    int8_t,
    int16_t,
    int32_t,
    int64_t,
    uint8_t,
    uint16_t,
    uint32_t,
    uint64_t,
)

from pandas._libs.algos cimport TiebreakEnumType

# from torchqtm._libs.window.aggregations import skiplist_init

# All numeric types except complex
ctypedef fused numeric_t:
    int8_t
    int16_t
    int32_t
    int64_t

    uint8_t
    uint16_t
    uint32_t
    uint64_t

    float32_t
    float64_t


from numpy cimport (
    float32_t,
    float64_t,
    int64_t,
    ndarray,
)

cdef:
    float32_t MINfloat32 = np.NINF
    float64_t MINfloat64 = np.NINF

    float32_t MAXfloat32 = np.inf
    float64_t MAXfloat64 = np.inf

    float64_t NaN = <float64_t>np.NaN

ctypedef struct node_t:
    node_t ** next
    int *width
    double value
    int is_nil
    int levels
    int ref_count

ctypedef struct skiplist_t:
    node_t *head
    node_t ** tmp_chain
    int *tmp_steps
    int size
    int maxlevels


cdef bint isnan(float64_t x) nogil:
    return x != x

cdef bint notnan(float64_t x) nogil:
    return x == x

cdef bint signbit(float64_t x) nogil:
    if x < 0.0:
        return 1
    else:
        return 0


#
#
cdef inline numeric_t calc_mm(int64_t minp, Py_ssize_t nobs,
                              numeric_t value) nogil:
    cdef:
        numeric_t result

    if numeric_t in cython.floating:
        if nobs >= minp:
            result = value
        else:
            result = NaN
    else:
        result = value

    return result


cdef inline numeric_t init_mm(numeric_t ai, Py_ssize_t *nobs, bint is_max) nogil:

    if numeric_t in cython.floating:
        if ai == ai:
            nobs[0] = nobs[0] + 1
        elif is_max:
            if numeric_t == cython.float:
                ai = MINfloat32
            else:
                ai = MINfloat64
        else:
            if numeric_t == cython.float:
                ai = MAXfloat32
            else:
                ai = MAXfloat64

    else:
        nobs[0] = nobs[0] + 1

    return ai

cdef inline void remove_mm(numeric_t aold, Py_ssize_t *nobs) nogil:
    """ remove a value from the mm calc """
    if numeric_t in cython.floating and aold == aold:
        nobs[0] = nobs[0] - 1


cdef inline float64_t calc_sum(int64_t minp, int64_t nobs, float64_t sum_x) nogil:
    cdef:
        float64_t result

    if nobs == 0 == minp:
        result = 0
    elif nobs >= minp:
        result = sum_x
    else:
        result = NaN

    return result


cdef inline void add_sum(float64_t val, int64_t *nobs, float64_t *sum_x,
                         float64_t *compensation) nogil:
    """ add a value from the sum calc using Kahan summation """

    cdef:
        float64_t y, t

    # Not NaN
    if notnan(val):
        nobs[0] = nobs[0] + 1
        y = val - compensation[0]
        t = sum_x[0] + y
        compensation[0] = t - sum_x[0] - y
        sum_x[0] = t


cdef inline void remove_sum(float64_t val, int64_t *nobs, float64_t *sum_x,
                            float64_t *compensation) nogil:
    """ remove a value from the sum calc using Kahan summation """

    cdef:
        float64_t y, t

    # Not NaN
    if notnan(val):
        nobs[0] = nobs[0] - 1
        y = - val - compensation[0]
        t = sum_x[0] + y
        compensation[0] = t - sum_x[0] - y
        sum_x[0] = t


def roll_sum(const float64_t[:] values, ndarray[int64_t] start,
             ndarray[int64_t] end, int64_t minp) -> np.ndarray:
    cdef:
        Py_ssize_t i, j
        float64_t sum_x, compensation_add, compensation_remove
        int64_t s, e
        int64_t nobs = 0, N = len(start)
        ndarray[float64_t] output
        bint is_monotonic_increasing_bounds

    is_monotonic_increasing_bounds = True

    output = np.empty(N, dtype=np.float64)

    with nogil:

        for i in range(0, N):
            s = start[i]
            e = end[i]

            if i == 0 or not is_monotonic_increasing_bounds or s >= end[i - 1]:

                # setup

                sum_x = compensation_add = compensation_remove = 0
                nobs = 0
                for j in range(s, e):
                    add_sum(values[j], &nobs, &sum_x, &compensation_add)

            else:

                # calculate deletes
                for j in range(start[i - 1], s):
                    remove_sum(values[j], &nobs, &sum_x, &compensation_remove)

                # calculate adds
                for j in range(end[i - 1], e):
                    add_sum(values[j], &nobs, &sum_x, &compensation_add)

            output[i] = calc_sum(minp, nobs, sum_x)

            if not is_monotonic_increasing_bounds:
                nobs = 0
                sum_x = 0.0
                compensation_remove = 0.0

    return output


def roll_max(ndarray[float64_t] values, ndarray[int64_t] start,
             ndarray[int64_t] end, int64_t minp) -> np.ndarray:
    """
    Moving max of 1d array of any numeric type along axis=0 ignoring NaNs.

    Parameters
    ----------
    values : np.ndarray[np.float64]
    window : int, size of rolling window
    minp : if number of observations in window
          is below this, output a NaN
    index : ndarray, optional
       index for window computation
    closed : 'right', 'left', 'both', 'neither'
            make the interval closed on the right, left,
            both or neither endpoints

    Returns
    -------
    np.ndarray[float]
    """
    return _roll_min_max(values, start, end, minp, is_max=1)


def roll_min(ndarray[float64_t] values, ndarray[int64_t] start,
             ndarray[int64_t] end, int64_t minp) -> np.ndarray:
    """
    Moving min of 1d array of any numeric type along axis=0 ignoring NaNs.

    Parameters
    ----------
    values : np.ndarray[np.float64]
    window : int, size of rolling window
    minp : if number of observations in window
          is below this, output a NaN
    index : ndarray, optional
       index for window computation

    Returns
    -------
    np.ndarray[float]
    """
    return _roll_min_max(values, start, end, minp, is_max=0)


cdef _roll_min_max(ndarray[numeric_t] values,
                   ndarray[int64_t] starti,
                   ndarray[int64_t] endi,
                   int64_t minp,
                   bint is_max):
    cdef:
        numeric_t ai
        int64_t curr_win_size, start
        Py_ssize_t i, k, nobs = 0, N = len(starti)
        deque Q[int64_t]  # min/max always the front
        deque W[int64_t]  # track the whole window for nobs compute
        ndarray[float64_t, ndim=1] output

    output = np.empty(N, dtype=np.float64)
    Q = deque[int64_t]()
    W = deque[int64_t]()

    with nogil:

        # This is using a modified version of the C++ code in this
        # SO post: https://stackoverflow.com/a/12239580
        # The original impl didn't deal with variable window sizes
        # So the code was optimized for that

        # first window's size
        curr_win_size = endi[0] - starti[0]
        # GH 32865
        # Anchor output index to values index to provide custom
        # BaseIndexer support
        for i in range(N):

            curr_win_size = endi[i] - starti[i]
            if i == 0:
                start = starti[i]
            else:
                start = endi[i - 1]

            for k in range(start, endi[i]):
                ai = init_mm(values[k], &nobs, is_max)
                # Discard previous entries if we find new min or max
                if is_max:
                    while not Q.empty() and ((ai >= values[Q.back()]) or
                                             values[Q.back()] != values[Q.back()]):
                        Q.pop_back()
                else:
                    while not Q.empty() and ((ai <= values[Q.back()]) or
                                             values[Q.back()] != values[Q.back()]):
                        Q.pop_back()
                Q.push_back(k)
                W.push_back(k)

            # Discard entries outside and left of current window
            while not Q.empty() and Q.front() <= starti[i] - 1:
                Q.pop_front()
            while not W.empty() and W.front() <= starti[i] - 1:
                remove_mm(values[W.front()], &nobs)
                W.pop_front()

            # Save output based on index in input value array
            if not Q.empty() and curr_win_size > 0:
                output[i] = calc_mm(minp, nobs, values[Q.front()])
            else:
                output[i] = NaN

    return output


cdef inline float64_t calc_mean(int64_t minp, Py_ssize_t nobs,
                                Py_ssize_t neg_ct, float64_t sum_x) nogil:
    cdef:
        float64_t result

    if nobs >= minp and nobs > 0:
        result = sum_x / <float64_t>nobs
        if neg_ct == 0 and result < 0:
            # all positive
            result = 0
        elif neg_ct == nobs and result > 0:
            # all negative
            result = 0
        else:
            pass
    else:
        result = NaN
    return result


cdef inline void add_mean(float64_t val, Py_ssize_t *nobs, float64_t *sum_x,
                          Py_ssize_t *neg_ct, float64_t *compensation) nogil:
    """ add a value from the mean calc using Kahan summation """
    cdef:
        float64_t y, t

    # Not NaN
    if notnan(val):
        nobs[0] = nobs[0] + 1
        y = val - compensation[0]
        t = sum_x[0] + y
        compensation[0] = t - sum_x[0] - y
        sum_x[0] = t
        if signbit(val):
            neg_ct[0] = neg_ct[0] + 1


cdef inline void remove_mean(float64_t val, Py_ssize_t *nobs, float64_t *sum_x,
                             Py_ssize_t *neg_ct, float64_t *compensation) nogil:
    """ remove a value from the mean calc using Kahan summation """
    cdef:
        float64_t y, t

    if notnan(val):
        nobs[0] = nobs[0] - 1
        y = - val - compensation[0]
        t = sum_x[0] + y
        compensation[0] = t - sum_x[0] - y
        sum_x[0] = t
        if signbit(val):
            neg_ct[0] = neg_ct[0] - 1


@cython.boundscheck(False)
@cython.wraparound(False)
def roll_mean(const float64_t[:] values, ndarray[int64_t] start,
              ndarray[int64_t] end, int64_t minp) -> np.ndarray:
    cdef:
        float64_t val, compensation_add, compensation_remove, sum_x
        int64_t s, e
        Py_ssize_t nobs, i, j, neg_ct, N = len(start)
        ndarray[float64_t] output
        bint is_monotonic_increasing_bounds

    is_monotonic_increasing_bounds = True

    output = np.empty(N, dtype=np.float64)

    with nogil:

        for i in range(0, N):
            s = start[i]
            e = end[i]

            if i == 0 or not is_monotonic_increasing_bounds or s >= end[i - 1]:

                compensation_add = compensation_remove = sum_x = 0
                nobs = neg_ct = 0
                # setup
                for j in range(s, e):
                    val = values[j]
                    add_mean(val, &nobs, &sum_x, &neg_ct, &compensation_add)

            else:

                # calculate deletes
                for j in range(start[i - 1], s):
                    val = values[j]
                    remove_mean(val, &nobs, &sum_x, &neg_ct, &compensation_remove)

                # calculate adds
                for j in range(end[i - 1], e):
                    val = values[j]
                    add_mean(val, &nobs, &sum_x, &neg_ct, &compensation_add)

            output[i] = calc_mean(minp, nobs, neg_ct, sum_x)

            if not is_monotonic_increasing_bounds:
                nobs = 0
                neg_ct = 0
                sum_x = 0.0
                compensation_remove = 0.0
    return output


@cython.boundscheck(False)
@cython.wraparound(False)
def roll_apply_1D(object obj,
                  ndarray[int64_t] start,
                  ndarray[int64_t] end,
                  int64_t minp,
                  object function,
                  tuple args,
                  dict kwargs) -> np.ndarray:
    cdef:
        ndarray[float64_t] output, counts
        ndarray[float64_t, cast=True] arr
        Py_ssize_t i, s, e,
        Py_ssize_t N = len(start)
        Py_ssize_t n = len(obj)

    if n == 0:
        return np.array([], dtype=np.float64)

    arr = np.asarray(obj)

    # ndarray input
    if not arr.flags.c_contiguous:
        arr = arr.copy('C')

    counts = roll_sum(np.isfinite(arr).astype(float), start, end, minp)

    output = np.empty(N, dtype=np.float64)

    for i in range(N):

        s = start[i]
        e = end[i]

        if counts[i] >= minp:
                output[i] = function(arr[s:e], *args, **kwargs)
        else:
            output[i] = NaN

    return output


@cython.boundscheck(False)
@cython.wraparound(False)
def roll_apply_2D(object obj,
                  ndarray[int64_t] start,
                  ndarray[int64_t] end,
                  int64_t minp,
                  object function,
                  tuple args,
                  dict kwargs) -> np.ndarray:
    cdef:
        ndarray[float64_t, ndim=2] output
        ndarray[float64_t, ndim=1] counts
        ndarray[float64_t, ndim=2, cast=True] arr
        Py_ssize_t i, s, e,
        Py_ssize_t N = len(start)
        Py_ssize_t n = len(obj)
        Py_ssize_t n_col = obj.shape[1]

    # if n == 0:
    #     return np.array([], dtype=np.float64)

    arr = np.asarray(obj)

    # ndarray input
    if not arr.flags.c_contiguous:
        arr = arr.copy('C')

    counts = roll_sum(np.isfinite(arr[:, 0]).astype(float), start, end, minp)

    output = np.empty((N, n_col), dtype=np.float64)

    for i in range(N):

        s = start[i]
        e = end[i]

        if counts[i] >= minp:
            output[i] = function(arr[s:e], *args, **kwargs)
        else:
            output[i] = NaN

    return output


@cython.boundscheck(False)
@cython.wraparound(False)
def roll_apply_3D(object obj,
                  ndarray[int64_t] start,
                  ndarray[int64_t] end,
                  int64_t minp,
                  object function,
                  tuple args,
                  dict kwargs) -> Type[ndarray]:
    cdef:
        ndarray[float64_t, ndim=2] output
        ndarray[float64_t, ndim=1] counts
        ndarray[float64_t, ndim=3, cast=True] arr
        Py_ssize_t i, s, e,
        Py_ssize_t N = len(start)
        Py_ssize_t n = len(obj)
        Py_ssize_t n_stocks = obj.shape[1]
        Py_ssize_t n_features = obj.shape[2]
    # if n == 0:
    #     return np.array([], dtype=np.float64)

    arr = np.asarray(obj)
    # ndarray input
    if not arr.flags.c_contiguous:
        arr = arr.copy('C')

    counts = roll_sum(np.isfinite(arr[:, 0, 0]).astype(float), start, end, minp)
    output = np.empty((N, n_stocks), dtype=np.float64)
    for i in range(N):
        s = start[i]
        e = end[i]

        if counts[i] >= minp:
            output[i] = function(arr[s:e], *args, **kwargs)
        else:
            output[i] = NaN
    return output


@cython.boundscheck(False)
@cython.wraparound(False)
def roll_weighted_sum(
    const float64_t[:] values, const float64_t[:] weights, int minp
) -> np.ndaray:
    return _roll_weighted_sum_mean(values, weights, minp, avg=0)


@cython.boundscheck(False)
@cython.wraparound(False)
def roll_weighted_mean(
    const float64_t[:] values, const float64_t[:] weights, int minp
) -> np.ndaray:
    return _roll_weighted_sum_mean(values, weights, minp, avg=1)


cdef float64_t[:] _roll_weighted_sum_mean(const float64_t[:] values,
                                          const float64_t[:] weights,
                                          int minp, bint avg):
    """
    Assume len(weights) << len(values)
    """
    cdef:
        float64_t[:] output, tot_wgt, counts
        Py_ssize_t in_i, win_i, win_n, in_n
        float64_t val_in, val_win, c, w

    in_n = len(values)
    win_n = len(weights)

    output = np.zeros(in_n, dtype=np.float64)
    counts = np.zeros(in_n, dtype=np.float64)
    if avg:
        tot_wgt = np.zeros(in_n, dtype=np.float64)

    elif minp > in_n:
        minp = in_n + 1

    minp = max(minp, 1)

    with nogil:
        if avg:
            for win_i in range(win_n):
                val_win = weights[win_i]
                if val_win != val_win:
                    continue

                for in_i in range(in_n - (win_n - win_i) + 1):
                    val_in = values[in_i]
                    if val_in == val_in:
                        output[in_i + (win_n - win_i) - 1] += val_in * val_win
                        counts[in_i + (win_n - win_i) - 1] += 1
                        tot_wgt[in_i + (win_n - win_i) - 1] += val_win

            for in_i in range(in_n):
                c = counts[in_i]
                if c < minp:
                    output[in_i] = NaN
                else:
                    w = tot_wgt[in_i]
                    if w == 0:
                        output[in_i] = NaN
                    else:
                        output[in_i] /= tot_wgt[in_i]

        else:
            for win_i in range(win_n):
                val_win = weights[win_i]
                if val_win != val_win:
                    continue

                for in_i in range(in_n - (win_n - win_i) + 1):
                    val_in = values[in_i]

                    if val_in == val_in:
                        output[in_i + (win_n - win_i) - 1] += val_in * val_win
                        counts[in_i + (win_n - win_i) - 1] += 1

            for in_i in range(in_n):
                c = counts[in_i]
                if c < minp:
                    output[in_i] = NaN

    return output


from functools import lru_cache

from torchqtm.data.rq_data_source.adjust import convert_date_to_int
from torchqtm.utils.exchange_calendar import XSHGExchangeCalendar, DateIterator
from torchqtm.assets import Asset
import numpy as np
import pandas as pd
import typing

from torchqtm.types import (
    BASE_FIELDS,
    HISTORY_FREQUENCIES,
    ASSET_TYPE,
)

from torchqtm.data.rq_data_source.data_source import (
    ReadersTracker,
)

from torchqtm.data.rq_data_source.readers import StockDailyBarReader
from torchqtm.utils.datetime_utils import DateTimeManager, DateTimeMixin


def _find_position_of_dt(bars: np.ndarray, dt: pd.Timestamp):
    dt = np.uint64(convert_date_to_int(dt))
    pos = bars['datetime'].searchsorted(dt)
    if pos >= len(bars) or bars['datetime'][pos] != dt:
        return None
    return pos


def convert_dt_to_session_label(dt: pd.Timestamp):
    return pd.Timestamp(year=dt.year, month=dt.month, day=dt.day)


class DataPortal(DateTimeMixin):
    # TODO: add cache for quick access to history data
    def __init__(self,
                 data_source=ReadersTracker(),
                 data_frequency: typing.Literal["daily", "minute"] = "daily",
                 trading_calendar: XSHGExchangeCalendar = None,
                 datetime_manager: DateTimeManager = None,
                 restrictions=None):
        DateTimeMixin.__init__(self, datetime_manager)

        self._data_source = data_source
        self._datetime = None
        self.data_frequency = data_frequency
        self._views = {}
        self._daily_mode = (self.data_frequency == "daily")

        self._adjust_minutes = False
        if trading_calendar is None:
            self.trading_calendar = XSHGExchangeCalendar()
        else:
            self.trading_calendar = trading_calendar
        self._date_iterator = DateIterator(start_date=self.trading_calendar.start_date)
        # self._is_restricted = restrictions.is_restricted

    @property
    def current_session(self):
        return self._date_iterator.current_session

    def current(self, asset, filed):
        return self.get_scalar_asset_spot_value(asset, filed, self.current_dt, self.data_frequency)

    def _ensure_reader_aligned(self, reader):
        pass

    def _reindex_extra_aligned(self, reader):
        pass

    def _reindex_extra_source(self, dt, source_date_index):
        pass

    def handle_extra_source(self, source_df, sim_params):
        pass

    def _get_pricing_reader(self, data_frequency):
        pass

    def get_last_traded_dt(
            self,
            asset,
            dt,
            data_frequency,
    ):
        pass

    def _get_daily_spot_value(self, asset, field, session_label: pd.Timestamp) -> typing.Optional[float]:
        bars = self._data_source.get_bar(asset, session_label, "1d")
        if bars is None:
            return None
        elif field == "price":
            return float(bars["close"].values)
        else:
            return float(bars[field].values)

    def get_spot_value(
            self,
            assets: typing.Iterable[Asset],
            field: str,
            dt: pd.Timestamp,
            data_frequency: str,
    ):
        """
        >>>self.get_spot_value([asset], 'close', dt, 'daily')
        106.25
        """
        if data_frequency != 'daily':
            raise NotImplementedError

        if data_frequency == "daily":
            session_label = self.trading_calendar.convert_dt_to_session(dt)
            return [self._get_daily_spot_value(asset, field, session_label) for asset in assets]

    def get_scalar_asset_spot_value(
        self,
        asset: Asset,
        filed: str,
        dt: pd.Timestamp,
        data_frequency: typing.Literal["daily", "minute"],
    ):
        # """
        # >>>self.get_spot_value([asset], 'close', dt, 'daily')
        # 106.25
        # """
        if data_frequency != "daily":
            return NotImplementedError

        return self._get_daily_spot_value(asset, filed, self.trading_calendar.convert_dt_to_session(dt))

    def get_adjustments(self, assets: typing.List[Asset], dt: pd.Timestamp, perspective_dt):
        """
        perspective_dt : pd.Timestamp
            The timestamp from which the data is being viewed back from.
        >>>self.get_adjustments([asset], "close", dt, pd.Timestamp('20160102'))
        [0.9832470935131806]
        """
        pass

    def get_adjusted_value(
            self,
            asset: Asset,
            dt: pd.Timestamp,
            data_frequency: HISTORY_FREQUENCIES,
            spot_value: typing.Optional[float] = None,
    ):
        """
        >>>self.get_adjusted_value(asset, "close", dt, pd.Timestamp('20160102'), "daily")
        104.47000368577544
        """

    def _get_minute_spot_value(self, asset, column, dt, ffill=False):
        pass

    def _get_days_for_window(self, end_date, bar_count):
        """
        >>>self._get_days_for_window(dt, 5)
        DatetimeIndex(['2014-12-29', '2014-12-30', '2014-12-31', '2015-01-02',
                       '2015-01-05'],
                      dtype='datetime64[ns]', freq='C')
        """

    def _get_history_daily_window(
            self,
            assets,
            end_dt,
            bar_count,
            field,
            data_frequency,
    ):
        """
        >>>self._get_history_daily_window([asset], dt, 5, "close", "daily")
        Out[53]:
                    Equity(8 [AAPL])
        2014-12-29            113.91
        2014-12-30            112.52
        2014-12-31            110.38
        2015-01-02            109.33
        2015-01-05            106.25
        """
        bars = self._all_daily_bars_of(asset)
        bars[field]

    def _get_history_daily_window_data(
            self,
            assets,
            days_for_window,
            end_dt,

    ):
        """
        这个应该用不太上了
        """
        pass

    def handle_minute_history_out_of_bounds(self):
        pass

    def _get_history_minute_window(self):
        pass

    def get_history_window(
            self,
            assets: typing.List[Asset],
            end_dt: pd.Timestamp,
            bar_count: int,
            frequency: typing.Literal["1d", "1m"],
            field: BASE_FIELDS,
            ffill: bool = False,
    ) -> pd.DataFrame:
        """
        不太明白
        """
    def history_bars(
            self,
            asset,
            fields,
            bar_count,
            frequency,
    ):
        if isinstance(fields, str):
            return self._data_source._history_bars_single(
                asset=asset,
                field=fields,
                bar_count=bar_count,
                frequency=frequency,
                dt=self.current_dt,
            )
        elif isinstance(fields, typing.Iterable):
            return self._data_source._history_bars_many(
                asset=asset,
                fields=fields,
                bar_count=bar_count,
                frequency=frequency,
                dt=self.current_dt,
            )

    def _get_minute_window_data(self, assets, field, minutes_for_window):
        pass

    def _get_daily_window_data(self, assets, field, days_in_window, extra_slot=True):
        pass

    def _get_adjustment_list(self, asset, adjustments_dict, table_name):
        pass

    def get_splits(self, assets, dt):
        """
        """
        return None

    def get_stock_dividends(self, sid, trading_days):
        pass




from .exchange_calendar import XSHGExchangeCalendar


def get_calendar(*args, **kwargs):
    if args[0] == "XSHG":
        return XSHGExchangeCalendar()
    else:
        return ValueError("Invalid calendar argument")




import tushare as ts
import pandas as pd
import numpy as np
from torchqtm.config import __TS_API__
pro = ts.pro_api(__TS_API__)


class BenchMark(object):
    def __init__(self, index_code, start_date, end_date):
        """
        index_code: '000905.SH', '399300.SZ'
        """
        self.index_code = index_code
        self.start_date = start_date
        self.end_date = end_date
        self.data = self._get_index_data()

    def _get_index_data(self):
        df = pro.index_daily(ts_code=self.index_code, start_date=self.start_date, end_date=self.end_date)
        df.sort_values('trade_date', inplace=True)
        df = df.set_index('trade_date')
        df.index = pd.to_datetime(df.index)
        del df['ts_code']
        df.columns = ['Close', 'Open', 'High', 'Low', 'PreClose', 'Change', 'PctChange', 'Volume', 'Amount']
        df['PctChange'] = df['PctChange'] / 100
        return df


if __name__ == '__main__':
    benchmark = BenchMark('000905.SH', '20180101', '20230101')


import codecs
import json
import locale
import os
import sys
from copy import copy
from itertools import chain
from typing import Dict, Iterable, Optional, Literal

import h5py
import numpy as np
import pandas as pd
from contextlib import contextmanager
from abc import ABCMeta, abstractmethod

from functools import lru_cache

from torchqtm.data.bundle.rqalpha import DEFAULT_PATH


TERM_ALLOWED = Literal['0S', '1M', '2M', '3M', '6M', '9M', '1Y', '2Y', '3Y', '4Y', '5Y', '6Y', '7Y',
                        '8Y', '9Y', '10Y', '15Y', '20Y', '30Y', '40Y', '50Y']


def convert_date_to_int(dt):
    t = dt.year * 10000 + dt.month * 100 + dt.day
    t *= 1e6
    return t


def convert_dates_to_int(dates: pd.DatetimeIndex):
    intDates = dates.year * 10000 + dates.month * 100 + dates.day
    intDates *= 1e6
    return intDates.values.astype(np.uint64)


class TradingCalendarReader(object):
    def __init__(self, path=None):
        if path is None:
            self.path = os.path.join(DEFAULT_PATH, "trading_dates.npy")
        else:
            self.path = path

    def get_trading_calendar(self) -> pd.DatetimeIndex:
        return pd.DatetimeIndex(pd.Series(np.load(self.path, allow_pickle=False), dtype=str))


# TODO: implement it
class FutureInfoReader(object):
    def __init__(self, path=None, custom_future_info=None):
        if path is None:
            self.path = os.path.join(DEFAULT_PATH, "future_info.json")
        else:
            self.path = path
        with open(self.path, "r") as json_file:
            self._default_data = {
                item.get("order_book_id") or item.get("underlying_symbol"): self._process_future_info_item(
                    item
                ) for item in json.load(json_file)
            }
        self._custom_data = custom_future_info
        self._future_info = {}

    @classmethod
    def _process_future_info_item(cls, item):
        item["commission_type"] = item["commission_type"]
        return item

    def get_future_info(self, instrument):
        order_book_id = instrument.order_book_id
        try:
            return self._future_info[order_book_id]
        except KeyError:
            custom_info = self._custom_data.get(order_book_id) or self._custom_data.get(instrument.underlying_symbol)
            info = self._default_data.get(order_book_id) or self._default_data.get(instrument.underlying_symbol)
            if custom_info:
                info = copy(info) or {}
                info.update(custom_info)
            elif not info:
                raise NotImplementedError(_("unsupported future instrument {}").format(order_book_id))
            return self._future_info.setdefault(order_book_id, info)


class InstrumentStore(object):
    def __init__(self, instruments, instrument_type):
        # type: (Iterable[Instrument], INSTRUMENT_TYPE) -> None
        self._instrument_type = instrument_type
        self._instruments = {}
        self._sym_id_map = {}

        for ins in instruments:
            if ins.type != instrument_type:
                continue
            self._instruments[ins.order_book_id] = ins
            self._sym_id_map[ins.symbol] = ins.order_book_id

    @property
    def instrument_type(self):
        # type: () -> INSTRUMENT_TYPE
        return self._instrument_type

    @property
    def all_id_and_syms(self):
        # type: () -> Iterable[str]
        return chain(self._instruments.keys(), self._sym_id_map.keys())

    def get_instruments(self, id_or_syms):
        # type: (Optional[Iterable[str]]) -> Iterable[Instrument]
        if id_or_syms is None:
            return self._instruments.values()
        order_book_ids = set()
        for i in id_or_syms:
            if i in self._instruments:
                order_book_ids.add(i)
            elif i in self._sym_id_map:
                order_book_ids.add(self._sym_id_map[i])
        return (self._instruments[i] for i in order_book_ids)


class ShareTransformationStore(object):
    def __init__(self, f):
        with codecs.open(f, 'r', encoding="utf-8") as store:
            self._share_transformation = json.load(store)

    def get_share_transformation(self, order_book_id):
        try:
            transformation_data = self._share_transformation[order_book_id]
        except KeyError:
            return
        return transformation_data["successor"], transformation_data["share_conversion_ratio"]


def _file_path(path):
    # why do this? non-ascii path in windows!!
    if sys.platform == "win32":
        try:
            l = locale.getlocale(locale.LC_ALL)[1]
        except TypeError:
            l = None
        if l and l.lower() == "utf-8":
            return path.encode("utf-8")
    return path


class h5_file:
    def __init__(self, path, *args, mode="r", **kwargs):
        self.path = path
        self.args = args
        self.mode = mode
        self.kwargs = kwargs

    def __enter__(self):
        try:
            self.h5 = h5py.File(self.path, *self.args, mode=self.mode, **self.kwargs)
        except OSError as e:
            raise RuntimeError(
                "open data bundle failed, you can remove {} and try to regenerate bundle: {}".format(self.path, e))
        return self.h5

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.h5.close()


class StockDailyBarReader(object):
    DEFAULT_DTYPES = np.dtype([
        ('datetime', np.uint64),
        ('open', np.float64),
        ('close', np.float64),
        ('high', np.float64),
        ('low', np.float64),
        ('volume', np.float64),
    ])

    ATTRIBUTES = ["datetime", "open", "close", "high", "low", "limit_up", "limit_down", "volume", "total_turnover"]

    ATTRIBUTE_TO_INDEX = {
        "datetime": 0,
        "open": 1,
        "close": 2,
        "high": 3,
        "low": 4,
        "volume": 5,
    }

    def __init__(self, path):
        if path is None:
            self._path = os.path.join(DEFAULT_PATH, "stocks.h5")
        else:
            self._path = path

    def load_data(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                return h5[symbol][:]
            except KeyError:
                return np.empty(0, dtype=self.DEFAULT_DTYPES)

    def get_date_range(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                data = h5[symbol]
                return data[0]['datetime'], data[-1]['datetime']
            except KeyError:
                return 20050104 * 1e6, 20050104 * 1e6

    def symbols(self):
        with h5_file(self._path) as h5:
            return list(h5.keys())


class FutureDailyBarReader(object):
    DEFAULT_DTYPES = np.dtype([
        ('datetime', np.uint64),
        ('open', np.float64),
        ('close', np.float64),
        ('high', np.float64),
        ('low', np.float64),
        ('volume', np.float64),
    ])

    def __init__(self, path=None):
        if path is None:
            self._path = os.path.join(DEFAULT_PATH, "futures.h5")
        else:
            self._path = path

    def load_data(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                return h5[symbol][:]
            except KeyError:
                return np.empty(0, dtype=self.DEFAULT_DTYPES)

    def get_date_range(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                data = h5[symbol]
                return data[0]['datetime'], data[-1]['datetime']
            except KeyError:
                return 20050104 * 1e6, 20050104 * 1e6

    def symbols(self):
        with h5_file(self._path) as h5:
            return list(h5.keys())


class FundDailyBarReader(object):
    DEFAULT_DTYPES = np.dtype([
        ('datetime', np.uint64),
        ('open', np.float64),
        ('close', np.float64),
        ('high', np.float64),
        ('low', np.float64),
        ('volume', np.float64),
    ])

    def __init__(self, path=None):
        if path is None:
            self._path = os.path.join(DEFAULT_PATH, "funds.h5")
        else:
            self._path = path

    def load_data(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                return h5[symbol][:]
            except KeyError:
                return np.empty(0, dtype=self.DEFAULT_DTYPES)

    def get_date_range(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                data = h5[symbol]
                return data[0]['datetime'], data[-1]['datetime']
            except KeyError:
                return 20050104 * 1e6, 20050104 * 1e6

    def symbols(self):
        with h5_file(self._path) as h5:
            return list(h5.keys())


class IndexDailyBarReader(object):
    DEFAULT_DTYPES = np.dtype([
        ('datetime', np.uint64),
        ('open', np.float64),
        ('close', np.float64),
        ('high', np.float64),
        ('low', np.float64),
        ('volume', np.float64),
    ])

    def __init__(self, path=None):
        if path is None:
            self._path = os.path.join(DEFAULT_PATH, "indexes.h5")
        else:
            self._path = path

    def load_data(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                return h5[symbol][:]
            except KeyError:
                return np.empty(0, dtype=self.DEFAULT_DTYPES)

    def get_date_range(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                data = h5[symbol]
                return data[0]['datetime'], data[-1]['datetime']
            except KeyError:
                return 20050104 * 1e6, 20050104 * 1e6

    def symbols(self):
        with h5_file(self._path) as h5:
            return list(h5.keys())


class DividendsReader(object):
    DEFAULT_DTYPES = np.dtype([
        ('book_closure_date', np.uint64),
        ('announcement_date', np.float64),
        ('dividend_cash_before_tax', np.float64),
        ('ex_dividend_date', np.uint64),
        ('payable_date', np.uint64),
        ('round_lot', np.float64)
    ])

    def __init__(self, path=None):
        if path is None:
            self._path = os.path.join(DEFAULT_PATH, "dividends.h5")
        else:
            self._path = path

    def load_data(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                return h5[symbol][:]
            except KeyError:
                return None


class SplitsReader(object):
    DEFAULT_DTYPES = np.dtype([
        ("ex_date", np.uint64),
        ("split_factor", np.float64),
    ])

    def __init__(self, path=None):
        if path is None:
            self._path = os.path.join(DEFAULT_PATH, "split_factor.h5")
        else:
            self._path = path

    def load_data(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                return h5[symbol][:]
            except KeyError:
                return None


class YieldCurveReader(object):
    def __init__(self, path=None):
        if path is None:
            self._path = os.path.join(DEFAULT_PATH, "yield_curve.h5")
        else:
            self._path = path
        self._data = None

    def load_data(
            self,
            start_date: pd.Timestamp,
            end_date: pd.Timestamp,
            term: Optional[TERM_ALLOWED] = None,
    ):
        with h5_file(self._path) as h5:
            self._data = h5["data"][:]

        d1 = int(convert_date_to_int(start_date) / 1e6)
        d2 = int(convert_date_to_int(end_date) / 1e6)

        s = self._data['date'].searchsorted(d1)
        e = self._data['date'].searchsorted(d2, side='right')

        if e == len(self._data):
            e -= 1
        if self._data[e]['date'] == d2:
            e += 1

        if e < s:
            return None

        df = pd.DataFrame(self._data[s:e])
        df.index = pd.to_datetime([str(d) for d in df['date']])
        del df['date']

        if term is not None:
            return df[term]
        return df


class ExCumFactorReader(object):
    def __init__(self, path=None):
        if path is None:
            self._path = os.path.join(DEFAULT_PATH, "ex_cum_factor.h5")
        else:
            self._path = path

    @lru_cache(None)
    def load_data(self, symbol: str):
        with h5_file(self._path) as h5:
            try:
                return h5[symbol][:]
            except KeyError:
                return None


class DatesReader(object):
    def __init__(self, path=None):
        if path is None:
            self._path = os.path.join(DEFAULT_PATH, "trading_dates.npy")
        else:
            self._path = path

    @lru_cache(None)
    def get_days(self, order_book_id):
        with h5_file(self._f) as h5:
            try:
                days = h5[order_book_id][:]
            except KeyError:
                return set()
            return set(days.tolist())

    def contains(self, symbol, dates):
        date_set = self.get_days(symbol)
        if not date_set:
            return None

        def _to_dt_int(d):
            if isinstance(d, (int, np.int64, np.uint64)):
                return int(d // 1000000) if d > 100000000 else int(d)
            else:
                return d.year * 10000 + d.month * 100 + d.day

        return [(_to_dt_int(d) in date_set) for d in dates]


if __name__ == "__main__":
    bar_stock = StockDailyBarReader(os.path.join(DEFAULT_PATH, 'stocks.h5'))
    bar_future = FutureDailyBarReader(os.path.join(DEFAULT_PATH, 'futures.h5'))
    split_reader = SplitsReader()
    calender = TradingCalendarReader()
    calender.get_trading_calendar()
    bar_stock.load_data("000001.XSHE")
    yield_curve = YieldCurveReader()
    yield_curve.load_data(pd.Timestamp('20100101'), pd.Timestamp('20150101'))
    bar_stock.get_date_range("000001.XSHE")
    bar_future.load_data("IF2101")

    # with h5_file(os.path.join(DEFAULT_PATH, 'futures.h5'), mode='r') as f:
    #     f["000001.XSHE"][:]





from torchqtm.edbt.temp.account import VirtualAccount
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from torchqtm.edbt.temp.datahandler import BarData
from mycalendar import Calendar
import pandas as pd


class BacktestEngine(object, metaclass=ABCMeta):

    def __init__(self, bars: BarData, accounts_names: Iterable[str]):
        self.bars = bars  # 想继承bars的一些方法
        self.accounts = {key: VirtualAccount(key, self.bars) for key in accounts_names}
        self.history = self.bars.history;
        self.calendar = Calendar(start_date=self.bars.start_date, end_date=self.bars.end_date)
        self.rebalance_date = self.create_rebalance_date()

    @property
    def now(self) -> pd.Timestamp:
        return self.bars.now

    @property
    def current_bar(self) -> int:
        return self.bars.bar_index

    def get_account(self, account_name):
        return self.accounts[account_name]

    def create_rebalance_date(self) -> list[pd.Timestamp]:
        temp = self.calendar.create_monthly_groups()  # TODO: 加入参数
        rlt = []
        for x in temp.values():
            rlt.append(x[-1])
        return rlt

    @abstractmethod
    def initialize(self):
        """
        """
        raise NotImplementedError('Should be implemented in subclass')

    @abstractmethod
    def handle_data(self):
        """
        """
        raise NotImplementedError('Should be implemented in subclass')

    # @abstractmethod
    # def generate_alpha_factor(self):
    #     """
    #     """
    #     raise NotImplementedError('Should be implemented in subclass')

    def run_strategy(self):
        self.initialize()

        while True:
            self.bars.update_bars()
            if not self.bars.continue_backtest:  # 没有取出数据的话则停止回测
                break
            self.handle_data()
            for account in self.accounts:
                self.accounts[account].update_account()

    def run_stats(self) -> dict:
        # TODO: 新增一个类, 用来计算各种指标
        rlt = {}
        for key in self.accounts:
            rlt[key] = self.accounts[key].calculate_returns()
        return rlt


class QuickBacktestEngine(object, metaclass=ABCMeta):

    pass


import pandas as pd
import numpy as np
from torchqtm.finance.positions.tracker import PositionsTracker
import typing


class Portfolio:
    """object providing information about current portfolio state

    Parameters
    ----------
    start_date : pd.Timestamp
        The start date for the period being recorded.
    capital_base : float
        The starting value for the portfolio. This will be used as the starting
        cash, current cash, and portfolio value.

    Attributes
    ----------
    positions : zipline.protocol.Positions
        Dict-like object containing information about currently-held positions.
    cash : float
        Amount of cash currently held in portfolio.
    portfolio_value : float
        Current liquidation value of the portfolio's holdings.
        This is equal to ``cash + sum(shares * price)``
    starting_cash : float
        Amount of cash in the portfolio at the start of the backtest.
    """
    def __init__(
            self,
            start_date: pd.Timestamp,
            capital_base: float = 0.0,
    ):
        self.cash_flow: float = 0.0
        self.starting_cash: float = capital_base
        self.portfolio_value: float = capital_base
        self.pnl: float = 0.0
        self.returns: float = 0.0
        self.cash: float = 0.0
        self.positions: typing.Dict = {}
        self.start_date: pd.Timestamp = start_date
        self.positions_value: float = 0.0
        self.positions_exposure: float = 0.0

    @property
    def capital_used(self):
        return self.cash_flow

    def __setattr__(self, key, value):
        raise AttributeError

    def __repr__(self):
        return "Portfolio({0})".format(self.__dict__)

    @property
    def current_portfolio_weights(self):
        """
        Compute each asset's weight in the portfolio by calculating its held
        value divided by the total value of all positions

        Each equity's value is its price times the number of shares held. Each
        futures contract's value is its unit price times number of shares held
        times the multiplier.
        """

        position_values = pd.Series(
            {
                asset: (
                        position.last_sale_price * position.amount * asset.price_multiplier
                )
                for asset, position in self.positions.items()
            },
            dtype=np.float64,
        )
        return position_values / self.portfolio_value











# distutils: language = c

import numpy as np
import pandas as pd
cimport cython
from libc.stdlib cimport malloc, free
from libc.string cimport memset

@cython.wraparound(False)
@cython.boundscheck(False)
def _group(x, int[::1] group, agg_func):
    if np.sum(np.isnan(x))+np.sum(np.isnan(group)):
        nan_mask = np.isnan(x) | np.isnan(group)

        x_copy = np.where(nan_mask, 0, x)
        group_copy = np.where(nan_mask, '0', group)

        labels, indices = np.unique(group_copy, return_inverse=True)
        grouped_val = np.array([agg_func(x_copy[group_copy == label]) for label in labels])
        rlt = grouped_val[indices]
        rlt = np.where(nan_mask, np.nan, rlt)
    else:
        labels, indices = np.unique(group, return_inverse=True)
        grouped_val = np.array([agg_func(x[group == label]) for label in labels])
        rlt = grouped_val[indices]
    return rlt

@cython.wraparound(False)
@cython.boundscheck(False)
def _group_neutralize_single(x, group):
    # Ensure x and group have same length
    assert len(x) == len(group), "Series x and group must have same length."

    # Calculate the mean of each group
    group_means = x.groupby(group).transform('mean')
    # Subtract the mean of each group from the corresponding elements in x
    neutralized_values = x - group.map(group_means)

    return neutralized_values


from collections import OrderedDict
import math
import logging
import numpy as np
import pandas as pd
import typing

from torchqtm.finance.orders.order import Order
from torchqtm.finance.portfolio import Portfolio
from torchqtm.finance.positions.tracker import PositionsTracker
from typing import Literal, List, Dict, Tuple, Optional, Any
from torchqtm.finance.positions.position import Position
from torchqtm.data.data_portal import DataPortal
from torchqtm.finance.transaction import Transaction
from torchqtm.types import DATA_FREQUENCIES, ASSET_TYPE, EVENT_TYPE
from torchqtm.assets import Asset
from torchqtm.finance.orders.tracker import OrdersTracker
from torchqtm.utils.datetime_utils import DateTimeManager, DateTimeMixin
log = logging.getLogger("Account")


class AccountTracker(object):
    pass


class Ledger:
    """Provide access to accounting infos
    """
    def __init__(
            self,
            start_date: pd.Timestamp,
            capital_base: float = 0.0,
            positions_tracker: PositionsTracker = None,
    ):
        self.cash_flow: float = 0.0
        self.starting_cash: float = capital_base
        self.portfolio_value: float = capital_base
        self.pnl: float = 0.0
        self.returns: float = 0.0
        self.cash: float = capital_base
        self.positions_tracker: PositionsTracker = positions_tracker
        self.start_date: pd.Timestamp = start_date
        self.positions_value: float = 0.0
        self.positions_exposure: float = 0.0

    @property
    def capital_used(self):
        return self.cash_flow

    @property
    def current_portfolio_weights(self):
        """
        Compute each asset's weight in the portfolio by calculating its held
        value divided by the total value of all positions.

        Each equity's value is its price times the number of shares held. Each
        futures contract's value is its unit price times number of shares held
        times the multiplier.
        """
        position_values = pd.Series(
            {
                asset: (
                    position.last_sale_price * position.amount * asset.price_multiplier
                )
                for asset, position in self.positions_tracker.data.items()
            },
            dtype=float,
        )
        return position_values / self.portfolio_value


class Account(DateTimeMixin):
    """
    The account trackers all get_orders and get_transactions as well as the current state of
    the portfolio and positions.
    """
    def __init__(
            self,
            datetime_manager: DateTimeManager,
            trading_sessions: pd.DatetimeIndex,
            capital_base: float,
            data_frequency: DATA_FREQUENCIES,
    ):
        DateTimeMixin.__init__(self, datetime_manager)

        if len(trading_sessions):
            start = trading_sessions[0]
        else:
            start = None

        # TODO: initialize
        self.positions_tracker: PositionsTracker = PositionsTracker(data_frequency)
        self.orders_tracker: OrdersTracker = OrdersTracker(
            datetime_manager=datetime_manager,
        )
        self.slippage_models = self.orders_tracker.slippage_models
        self.commission_models = self.orders_tracker.commission_models

        # ledger info
        self._ledger = Ledger(start, capital_base, self.positions_tracker)

        # 构建一个reference
        self.positions: typing.Dict = self.positions_tracker.data
        self.start_date: pd.Timestamp = start

        self._need_update_ledger = False
        self.daily_returns_series: pd.Series = pd.Series(np.nan, index=trading_sessions)

        self._previous_total_returns: float = 0.0

        self._position_stats: Optional[Any] = None

        self._processed_transactions_by_dt: Dict[pd.Timestamp, Transaction] = {}

        self._orders_by_dt: Dict[pd.Timestamp, OrderedDict] = {}

        self._orders_by_id: typing.OrderedDict[str, Order] = OrderedDict()

        self._payout_last_sale_prices: Dict[Asset, float] = {}

    @property
    def todays_returns(self):
        # compute today's returns in returns space instead of portfolio-value
        # space to work even when we have capital changes
        return (self.ledger.returns + 1) / (self._previous_total_returns + 1) - 1

    @property
    def ledger(self):
        self.update_ledger()
        return self._ledger

    def sync_last_sale_price(
            self,
            dt: pd.Timestamp,
            data_portal: DataPortal,
            handle_non_market_minutes: bool = False,
    ):
        self.positions_tracker.sync_last_sale_price(
            dt,
            data_portal,
            handle_non_market_minutes,
        )
        self._need_update_ledger = True

    @staticmethod
    def _calculate_payout(multiplier, amount, old_price, price):
        return (price - old_price) * multiplier * amount

    def update_cash_flow(self, amount: float):
        self._need_update_ledger = True
        self._ledger.cash_flow += amount
        self._ledger.cash += amount

    def handle_transaction(self, txn: Transaction):
        """Add a transaction to ledger, updating the current state as needed.

        Parameters
        ----------
        txn : Transaction
            The transaction to execute.
        """
        asset = txn.asset
        if asset.type == ASSET_TYPE.Future:
            try:
                old_price = self._payout_last_sale_prices[asset]
            except KeyError:
                self._payout_last_sale_prices[asset] = txn.price
            else:
                position = self.positions[asset]
                amount = position.amount
                price = txn.price

                self.update_cash_flow(
                    self._calculate_payout(
                        asset.price_multiplier,
                        amount,
                        old_price,
                        price,
                    ),
                )

                if amount + txn.amount == 0:
                    del self._payout_last_sale_prices[asset]
                else:
                    self._payout_last_sale_prices[asset] = price
        else:
            self.update_cash_flow(-(txn.price * txn.amount))

        self.positions_tracker.handle_transaction(txn)

        try:
            self._processed_transactions_by_dt[txn.dt].append(
                txn,
            )
        except KeyError:
            self._processed_transactions_by_dt[txn.dt] = [txn]

    def handle_splits(self, splits: List[Tuple[Asset, float]]):
        """Process a list of splits by modifying any positions as needed.

        Parameters
        ----------
        splits: List[Tuple[Asset, float]]
            A list of splits. Each split is a tuple of (asset, ratio)

        """
        leftover_cash = self.positions_tracker.handle_splits(splits)
        if leftover_cash > 0:
            self.update_cash_flow(leftover_cash)

    def handle_order(self, order: Order):
        """Keep track of an order that was placed.

        Parameters
        ----------
        order : Order
            The order to record.
        """
        try:
            dt_orders = self._orders_by_dt[order.dt]
        except KeyError:
            self._orders_by_dt[order.dt] = OrderedDict(
                [
                    (order.id, order),
                ]
            )
            self._orders_by_id[order.id] = order
        else:
            self._orders_by_id[order.id] = dt_orders[order.id] = order
            # to preserve the order of the get_orders by modified date
            OrderedDict.move_to_end(dt_orders, order.id, last=True)

        OrderedDict.move_to_end(self._orders_by_id, order.id, last=True)

    def handle_commission(self, commission):
        """Process the commission.

        Parameters
        ----------
        commission : zp.Event
            The commission being paid.
        """
        # TODO: fix this
        asset = commission["asset"]
        cost = commission["cost"]
        self.positions_tracker.handle_commission(asset, cost)
        # 为什么是-cost
        self.update_cash_flow(-cost)

    def close_position(self, asset, dt, data_portal):
        # TODO: fix this
        txn = self.positions_tracker.maybe_create_position_transaction(
            asset,
            dt,
            data_portal,
        )
        if txn is not None:
            self.handle_transaction(txn)

    def handle_dividends(self, next_session, asset_finder, adjustment_reader):
        # TODO: fix this
        pass

    def capital_change(self, change_amount: float) -> None:
        self.update_ledger()
        self._ledger.portfolio_value += change_amount
        self._ledger.cash += change_amount

    def get_transactions(self, dt: Optional[pd.Timestamp] = None):
        """Retrieve the dict-form of all of the get_transactions in a given bar or
        for the whole simulation.

        Parameters
        ----------
        dt : pd.Timestamp or None, optional
            The particular datetime to look up get_transactions for. If not passed,
            or None is explicitly passed, all of the get_transactions will be
            returned.

        Returns
        -------
        get_transactions : List[Transaction]
            The transaction information.
        """
        if dt is None:
            # flatten the by-day get_transactions
            return [
                txn
                for by_day in self._processed_transactions_by_dt.values()
                for txn in by_day
            ]

        return self._processed_transactions_by_dt.get(dt, [])

    def get_orders(self, dt=None):
        if dt is None:
            # get_orders by id is already flattened
            return [o.to_dict() for o in self._orders_by_id.values()]

        return [o.to_dict() for o in self._orders_by_dt.get(dt, {}).values()]

    def _get_payout_total(self, positions):
        calculate_payout = self._calculate_payout
        payout_last_sale_prices = self._payout_last_sale_prices

        total = 0
        for asset, old_price in payout_last_sale_prices.items():
            position = positions[asset]
            payout_last_sale_prices[asset] = position.last_sale_price
            price = position.last_sale_price
            amount = position.amount
            total += calculate_payout(
                asset.price_multiplier,
                amount,
                old_price,
                price,
            )

        return total

    def update_ledger(self):
        if not self._need_update_ledger:
            return None

        self._ledger.positions_value = self.positions_tracker.stats.net_value
        position_value = self.positions_tracker.stats.net_value
        self._ledger.positions_exposure = self.positions_tracker.stats.net_exposure
        self.update_cash_flow(self._get_payout_total(self.positions_tracker.data))

        start_value = self._ledger.portfolio_value

        # update the new starting value
        self._ledger.portfolio_value = self._ledger.cash + position_value
        end_value = self._ledger.cash + position_value

        pnl = end_value - start_value
        if start_value != 0:
            returns = pnl / start_value
        else:
            returns = 0.0

        self._ledger.pnl += pnl
        self._ledger.returns = (1 + self._ledger.returns) * (1 + returns) - 1

        # the portfolio has been fully synced
        self._need_update_ledger = False

    def subscribe_cash(self, amount: float):
        log.warning("有外部资金流入, 风险测度可能不再准确")
        self.update_cash_flow(amount)

    def get_portfolio(self):
        """Compute the current portfolio.

        Notes
        -----
        This is cached, repeated access will not recompute the portfolio until
        the portfolio may have changed.
        """
        # 这么搞一波相当于是每次调用portfolio的时候都会更新一次
        self.update_ledger()
        return self._portfolio

    def calculate_period_stats(self):
        self.update_ledger()
        if self._portfolio_value == 0:
            gross_leverage = net_leverage = np.inf
        else:
            gross_leverage = self.positions_tracker.stats.gross_exposure / self._portfolio_value
            net_leverage = self.positions_tracker.stats.net_exposure / self._portfolio_value
        return self._portfolio_value, gross_leverage, net_leverage

    # 提供orders_tracker的封装
    def set_dt(self, dt):
        return self.orders_tracker.set_dt(dt)

    def order(
            self,
            asset: Asset,
            amount: int,
            stop: float = None,
            limit: float = None,
            order_id: str = None,
    ) -> Optional[str]:
        return self.orders_tracker.order(asset, amount, stop, limit, order_id)

    def cancel(
            self,
            order_id: str,
            relay_status: bool = True,
    ) -> None:
        return self.orders_tracker.cancel(order_id, relay_status)

    def cancel_all_orders_for_asset(
            self,
            asset: Asset,
            warn: bool = False,
            relay_status: bool = True,
    ) -> None:
        return self.orders_tracker.cancel_all_orders_for_asset(asset, warn, relay_status)

    def execute_daily_cancel_policy(self, event: EVENT_TYPE) -> None:
        return self.orders_tracker.execute_daily_cancel_policy(event)

    def execute_cancel_policy(self, event: EVENT_TYPE) -> None:
        return self.orders_tracker.execute_cancel_policy(event)

    # Provide a consistent interface for metrics
    def start_of_session(self, session_label: pd.Timestamp):
        # We don not use the parameter session_label
        self._processed_transactions_by_dt.clear()
        self._orders_by_dt.clear()
        self._orders_by_id.clear()
        # 这里只是存了一波信息, 因为在start_of_session时我们似乎没有调用能够切换状态的函数
        self._previous_total_returns = self.ledger.returns

    def end_of_bar(self, session_idx):
        self.daily_returns_series.values[session_idx] = self.todays_returns

    def end_of_session(self, session_idx: int):
        self.daily_returns_series.values[session_idx] = self.todays_returns

    # def handle_start_of_session(
    #         self,
    #         session_label: pd.Timestamp,
    #         data_portal: DataPortal,
    # ):
    #     self.start_of_session(session_label)
    #     # self.process_dividends
    #     # self._current_session = session_label
    #
    # def handle_end_of_session(
    #         self,
    #         dt: pd.Timestamp,
    #         data_portal: DataPortal,
    # ):
    #     self.sync_last_sale_price(dt, data_portal)
    #     self.end_of_session()

    def get_position(self, asset: Asset):
        return self.positions_tracker.get_position(asset)



import tushare as ts
import pandas as pd
from torchqtm.config import __TS_API__
import mibian
pro = ts.pro_api(__TS_API__)
exchanges = ['SHFE', 'SH']
df = pro.opt_basic(exchange='CFFEX', fields='ts_code,name,exercise_type,list_date,delist_date')

ts.pro_bar(ts_code='000001.SZ', adj='qfq', start_date='20180101', end_date='20181011')


import typing
import builtins
import pandas as pd
import datetime
from enum import Enum

# from builtins import FuncType
Callable = typing.Callable
Any = typing.Any
TypeVar = typing.TypeVar

FuncType = Callable[..., Any]
F = TypeVar("F", bound=FuncType)
DocType = Callable[[FuncType], FuncType]  # 装饰器空间
D = TypeVar("D", bound=DocType)

# Callable[[...], DocType], 这个就是pandas中那个doc属于的空间

List = typing.List
Tuple = typing.Tuple
Dict = typing.Dict
Union = typing.Union
Optional = typing.Optional
Iterator = typing.Iterator
Sequence = typing.Sequence
Iterable = typing.Iterable
Literal = typing.Literal

BASE_FIELDS = typing.Literal['open', 'high', 'low', 'close', 'volume', 'price', 'contract', 'last_traded']
OHLCV_FIELDS = typing.Literal['open', 'high', 'low', 'close', 'volume']
OHLCVP_FIELDS = typing.Literal['open', 'high', 'low', 'close', 'volume', 'price']

# 这俩实在是太容易搞混了
HISTORY_FREQUENCIES = typing.Literal['1m', '1d']
DATA_FREQUENCIES = typing.Literal["minute", "daily"]


DateLike = typing.Union[pd.Timestamp, str, int, float, datetime.datetime]


class ORDER_STATUS(Enum):
    OPEN = 0
    FILLED = 1
    CANCELLED = 2
    REJECTED = 3
    HELD = 4


class ORDER_TYPE(Enum):
    MARKET = 0
    LIMIT = 1
    STOP = 2
    STOP_LIMIT = 3


class EVENT_TYPE(Enum):
    BAR = 0  # end of bar
    SESSION_START = 1  # start of a trading day
    SESSION_END = 2  # end of a trading day
    MINUTE_START = 3
    MINUTE_END = 4
    BEFORE_TRADING_START = 5  # start of trading
    MARKET_OPEN = 6


class ASSET_TYPE(Enum):
    Equity = 0
    Future = 1
    Index = 2
    Fund = 3


class TestError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self, message)









import torchqtm.op as op
import torchqtm.op.functional as F
from torchqtm.op.functional import *
import pandas as pd
import numpy as np
from abc import ABCMeta, abstractmethod
from torchqtm.tdbt.backtest import BackTestEnv
from torchqtm.base import BaseAlpha


class ChatGPT(BaseAlpha, metaclass=ABCMeta):
    def __init__(self, env: BackTestEnv, *args, **kwargs):
        super().__init__(env, *args, **kwargs)
        self.type = 'WorldQuant Alpha101'

    @abstractmethod
    def forward(self, *args, **kwargs):
        """assign self.rawdata and return"""
        raise NotImplementedError


class Alpha001(ChatGPT):
    def __init__(self, env):
        super().__init__(env)

    def forward(self):
        self.data = cs_rank(ts_sum(F.max(self.high - self.close, self.close - self.low), 5)) / 5
        return self.data


class Alpha002(ChatGPT):
    def __init__(self, env):
        super().__init__(env)
        self.Enabled = True

    def __repr__(self):
        return ""

    def forward(self):
        VWAP = ts_sum(self.volume * self.close, 20) / ts_sum(self.volume, 20)
        self.data = cs_rank(self.close - VWAP) / cs_rank(ts_std_dev(self.close, 20))
        return self.data


class Alpha003(ChatGPT):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return ""

    def forward(self):
        self.data = cs_rank(self.env.MktVal) * cs_rank(self.env.PE)
        return self.data


class Alpha004(ChatGPT):
    def __init__(self, env):
        super().__init__(env)

    def __repr__(self):
        return ""

    def forward(self):
        self.data = -cs_rank(ts_corr(cs_rank(self.close), cs_rank(ts_mean(self.volume, 10)), 6))
        return self.data
